
from gym.spaces import Discrete
import pytest

from abmarl.policies.policy import RandomPolicy


def test_policy_action_space():
    with pytest.raises(AssertionError):
        RandomPolicy(observation_space=Discrete(3))
    with pytest.raises(AssertionError):
        RandomPolicy(observation_space=Discrete(3), action_space={'move': Discrete(3)})
    RandomPolicy(observation_space=Discrete(3), action_space=Discrete(9))


def test_policy_observation_space():
    with pytest.raises(AssertionError):
        RandomPolicy(action_space=Discrete(3))
    with pytest.raises(AssertionError):
        RandomPolicy(action_space=Discrete(3), observation_space={'move': Discrete(3)})
    RandomPolicy(action_space=Discrete(3), observation_space=Discrete(9))


def test_compute_action():
    policy = RandomPolicy(action_space=Discrete(4), observation_space=Discrete(2))
    with pytest.raises(AssertionError):
        policy.compute_action(3)
    action = policy.compute_action(0)
    assert action in policy.action_space




from abmarl.sim.gridworld.examples.maze_navigation import MazeNavigationAgent, MazeNaviationSim
from abmarl.sim.gridworld.agent import GridWorldAgent
from abmarl.managers import AllStepManager
from abmarl.external import MultiAgentWrapper

object_registry = {
    'N': lambda n: MazeNavigationAgent(
        id='navigator',
        encoding=1,
        view_range=2,
        render_color='blue',
    ),
    'T': lambda n: GridWorldAgent(
        id='target',
        encoding=3,
        render_color='green'
    ),
    'W': lambda n: GridWorldAgent(
        id=f'wall{n}',
        encoding=2,
        blocking=True,
        render_shape='s'
    )
}

file_name = 'maze.txt'
sim = MultiAgentWrapper(
    AllStepManager(
        MazeNaviationSim.build_sim_from_file(
            file_name,
            object_registry,
            overlapping={1: [3], 3: [1]}
        )
    )
)


sim_name = "MazeNavigation"
from ray.tune.registry import register_env
register_env(sim_name, lambda sim_config: sim)


policies = {
    'navigator': (
        None,
        sim.sim.agents['navigator'].observation_space,
        sim.sim.agents['navigator'].action_space,
        {}
    )
}


def policy_mapping_fn(agent_id):
    return 'navigator'


# Experiment parameters
params = {
    'experiment': {
        'title': f'{sim_name}',
        'sim_creator': lambda config=None: sim,
    },
    'ray_tune': {
        'run_or_experiment': 'A2C',
        'checkpoint_freq': 50,
        'checkpoint_at_end': True,
        'stop': {
            'episodes_total': 2000,
        },
        'verbose': 2,
        'config': {
            # --- Simulation ---
            'disable_env_checking': False,
            'env': sim_name,
            'horizon': 200,
            'env_config': {},
            # --- Multiagent ---
            'multiagent': {
                'policies': policies,
                'policy_mapping_fn': policy_mapping_fn,
            },
            # "lr": 0.0001,
            # --- Parallelism ---
            # Number of workers per experiment: int
            "num_workers": 7,
            # Number of simulations that each worker starts: int
            "num_envs_per_worker": 1, # This must be 1 because we are not "threadsafe"
        },
    }
}



import fnmatch
import os


def custom_import_module(full_config_path):
    """
    Import and execute a python file as a module. Useful for import the experiment module and the
    analysis module.

    Args:
        full_config_path: Full path to the python file.

    Returns: The python file as a module
    """
    import importlib.util
    spec = importlib.util.spec_from_file_location("mod", full_config_path)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)
    return mod


def checkpoint_from_trained_directory(full_trained_directory, checkpoint_desired):
    """
    Return the checkpoint directory to load the policy. If checkpoint_desired is specified and
    found, then return that policy. Otherwise, return the last policy.
    """
    checkpoint_dirs = find_dirs_in_dir('checkpoint*', full_trained_directory)

    # Try to load the desired checkpoint
    if checkpoint_desired is not None: # checkpoint specified
        for checkpoint in checkpoint_dirs:
            if checkpoint_desired == int(checkpoint.split('/')[-1].split('_')[-1]):
                return checkpoint, checkpoint_desired
        import warnings
        warnings.warn(
            f'Could not find checkpoint_{checkpoint_desired}. Attempting to load the last '
            'checkpoint.'
        )

    # Load the last checkpoint
    max_checkpoint = None
    max_checkpoint_value = 0
    for checkpoint in checkpoint_dirs:
        checkpoint_value = int(checkpoint.split('/')[-1].split('_')[-1])
        if checkpoint_value > max_checkpoint_value:
            max_checkpoint_value = checkpoint_value
            max_checkpoint = checkpoint

    if max_checkpoint is None:
        raise FileNotFoundError("Did not find a checkpoint file in the given directory.")

    return max_checkpoint, max_checkpoint_value


def find_dirs_in_dir(pattern, path):
    """
    Traverse the path looking for directories that match the pattern.

    Return: list of paths that match
    """
    result = []
    for root, dirs, files in os.walk(path):
        for name in dirs:
            if fnmatch.fnmatch(name, pattern):
                result.append(os.path.join(root, name))
    return result




import numpy as np

from abmarl.trainers import SinglePolicyTrainer
from abmarl.tools import numpy_utils as npu


class OnPolicyMonteCarloTrainer(SinglePolicyTrainer):
    def train(self, iterations=10_000, gamma=0.9, **kwargs):
        state_action_returns = {}

        for i in range(iterations):
            states, actions, rewards = self.generate_episode(**kwargs)
            states = next(iter(states.values()))
            states.pop() # Pop off the terminating state.
            states = np.stack(states)
            actions = np.stack(next(iter(actions.values())))
            rewards = next(iter(rewards.values()))
            G = 0
            for i in reversed(range(len(states))):
                state, action, reward = states[i], actions[i], rewards[i]
                G = gamma * G + reward
                if not npu.array_in_array(state, states[:i]):
                    if not (npu.array_in_array(action, actions[:i])):
                        try:
                            state_action_returns[(state, action)].append(G)
                        except KeyError:
                            state_action_returns[(state, action)] = [G]
                        self.policy.update(
                            state, action, np.mean(state_action_returns[(state, action)])
                        )



def create_parser(subparsers):
    """Parse the arguments for the debug command.

    Returns:
        parser: ArgumentParser for debug command.
    """
    debug_parser = subparsers.add_parser(
        'debug',
        help="Run the simulation from the configruation file with random "
             "actions, rendering each step and outputting observations and actions. "
             "This is useful to ensure that you've setup the simulation correctly."
    )
    debug_parser.add_argument(
        'configuration', type=str, help='Path to python config file. Include the .py extension.'
    )
    debug_parser.add_argument(
        '-n', '--episodes', type=int, default=1, help='The number of episodes to run. Default 1.'
    )
    debug_parser.add_argument(
        '-s', '--steps-per-episode', type=int, default=200,
        help='The maximum number of steps to take per epsiode. Default 200.'
    )
    debug_parser.add_argument(
        '--render', action='store_true', help='Render the simulation each step.'
    )
    return debug_parser


def run(full_config_path, parameters):
    from abmarl import debug
    debug.run(full_config_path, parameters)




from abc import ABC, abstractmethod

from gym.spaces import Box
import numpy as np

from abmarl.sim.gridworld.base import GridWorldBaseComponent
from abmarl.sim.gridworld.agent import GridObservingAgent
import abmarl.sim.gridworld.utils as gu


class ObserverBaseComponent(GridWorldBaseComponent, ABC):
    """
    Abstract Observer Component base from which all observer components will inherit.
    """
    @property
    @abstractmethod
    def key(self):
        """
        The key in the observation dictionary.

        The observation space of all observing agents in the gridworld framework is a dict.
        We can build up complex observation spaces with multiple components by
        assigning each component an entry in the observation dictionary. Observations
        will be a dictionary even if your simulation only has one Observer.
        """
        pass

    @property
    @abstractmethod
    def supported_agent_type(self):
        """
        The type of Agent that this Observer works with.

        If an agent is this type, the Observer will add its entry to the
        agent's observation space and will produce observations for this agent.
        """
        pass

    @abstractmethod
    def get_obs(self, agent, **kwargs):
        """
        Observe the state of the simulation.

        Args:
            agent: The agent for which we return an observation.

        Returns:
            This agent's observation.
        """
        pass


class SingleGridObserver(ObserverBaseComponent):
    """
    Observe a subset of the grid centered on the agent's position.

    The observation is centered around the observing agent's position. Each agent
    in the "observation window" is recorded in the relative cell using its encoding.
    If there are multiple agents on a single cell with different encodings, the
    agent will observe only one of them chosen at random.
    """
    def __init__(self, observe_self=True, **kwargs):
        super().__init__(**kwargs)
        self.observe_self = observe_self
        max_encoding = max([agent.encoding for agent in self.agents.values()])
        for agent in self.agents.values():
            if isinstance(agent, self.supported_agent_type):
                agent.observation_space[self.key] = Box(
                    -2, max_encoding, (agent.view_range * 2 + 1, agent.view_range * 2 + 1), int
                )

    @property
    def key(self):
        """
        This Observer's key is "grid".
        """
        return 'grid'

    @property
    def supported_agent_type(self):
        """
        This Observer works with GridObservingAgents.
        """
        return GridObservingAgent

    @property
    def observe_self(self):
        """
        Agents can observe themselves, which may hide important information if
        overlapping is important. This can be turned off by setting observe_self
        to False.
        """
        return self._observe_self

    @observe_self.setter
    def observe_self(self, value):
        assert type(value) is bool, "Observe self must be a boolean."
        self._observe_self = value

    def get_obs(self, agent, **kwargs):
        """
        The agent observes a sub-grid centered on its position.

        The observation may include other agents, empty spaces, out of bounds, and
        masked cells, which can be blocked from view by other blocking agents.

        Returns:
            The observation as a dictionary.
        """
        if not isinstance(agent, self.supported_agent_type):
            return {}

        # Generate a local grid and an observation mask
        local_grid, mask = gu.create_grid_and_mask(
            agent, self.grid, agent.view_range, self.agents
        )

        # Convolve the grid observation with the mask.
        obs = np.zeros((2 * agent.view_range + 1, 2 * agent.view_range + 1), dtype=int)
        for r in range(2 * agent.view_range + 1):
            for c in range(2 * agent.view_range + 1):
                if mask[r, c]: # We can see this cell
                    candidate_agents = local_grid[r, c]
                    if candidate_agents is None: # This cell is out of bounds
                        obs[r, c] = -1
                    elif not candidate_agents: # In bounds empty cell
                        obs[r, c] = 0
                    else: # Observe one of the agents at this cell
                        if self.observe_self:
                            obs[r, c] = np.random.choice([
                                other.encoding for other in candidate_agents.values()
                            ])
                        else:
                            choices = [
                                other.encoding
                                for other in candidate_agents.values()
                                if other.id != agent.id
                            ]
                            # It may be that the observing agent is the only agent
                            # at this location but it cannot observe itself, which
                            # makes choices an empty list.
                            obs[r, c] = np.random.choice(choices) if choices else 0
                else: # Cell blocked by agent. Indicate invisible with -2
                    obs[r, c] = -2

        return {self.key: obs}


class MultiGridObserver(ObserverBaseComponent):
    """
    Observe a subset of the grid centered on the agent's position.

    The observation is centered around the observing agent's position. The observing
    agent sees a stack of observations, one for each positive encoding, where the
    number of agents of each encoding is given rather than the encoding
    itself. Out of bounds and masked indicators appear in every grid.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.number_of_encodings = max([agent.encoding for agent in self.agents.values()])
        for agent in self.agents.values():
            if isinstance(agent, self.supported_agent_type):
                agent.observation_space[self.key] = Box(
                    -2,
                    len(self.agents),
                    (agent.view_range * 2 + 1, agent.view_range * 2 + 1, self.number_of_encodings),
                    int
                )

    @property
    def key(self):
        """
        This Observer's key is "grid".
        """
        return 'grid'

    @property
    def supported_agent_type(self):
        """
        This Observer works with GridObservingAgents.
        """
        return GridObservingAgent

    def get_obs(self, agent, **kwargs):
        """
        The agent observes one or more sub-grids centered on its position.

        The observation may include other agents, empty spaces, out of bounds, and
        masked cells, which can be blocked from view by other blocking agents.
        Each grid records the number of agents on a particular cell correlated
        to a specific encoding.

        Returns:
            The observation as a dictionary.
        """
        if not isinstance(agent, self.supported_agent_type):
            return {}

        # Generate a local grid and an observation mask.
        local_grid, mask = gu.create_grid_and_mask(
            agent, self.grid, agent.view_range, self.agents
        )

        # Convolve the grid observation with the mask.
        obs = np.zeros(
            (2 * agent.view_range + 1, 2 * agent.view_range + 1, self.number_of_encodings),
            dtype=int
        )
        for encoding in range(self.number_of_encodings):
            for r in range(2 * agent.view_range + 1):
                for c in range(2 * agent.view_range + 1):
                    if mask[r, c]: # We can see this cell
                        candidate_agents = local_grid[r, c]
                        if candidate_agents is None: # This cell is out of bounds
                            obs[r, c, encoding] = -1
                        elif not candidate_agents: # In bounds empty cell
                            obs[r, c, encoding] = 0
                        else: # Observe the number of agents at this cell with this encoding
                            obs[r, c, encoding] = sum([
                                True if other.encoding == encoding + 1 else False
                                for other in candidate_agents.values()
                            ])
                    else: # Cell blocked by agent. Indicate invisible with -2
                        obs[r, c, encoding] = -2

        return {self.key: obs}




import numpy as np


class Grid:
    """
    A Grid stores the agents at indices in a numpy array.

    Components can interface with the Grid. Each index in the
    grid is a dictionary that maps the agent id to the agent object itself. If agents
    can overlap, then there may be more than one agent per cell.

    Args:
        rows: The number of rows in the grid.
        cols: The number of columns in the grid.
        overlapping: Dictionary that maps the agents' encodings to a list of encodings
            with which they can occupy the same cell. To avoid undefined behavior, the
            overlapping should be symmetric, so that if 2 can overlap with 3, then 3
            can also overlap with 2.
    """
    def __init__(self, rows, cols, overlapping=None, **kwargs):
        assert type(rows) is int and rows > 0, "Rows must be a positive integer."
        assert type(cols) is int and cols > 0, "Cols must be a positive integer."
        self._internal = np.empty((rows, cols), dtype=object)

        # Overlapping matrix
        if overlapping is not None:
            assert type(overlapping) is dict, "Overlap mapping must be dictionary."
            for k, v in overlapping.items():
                assert type(k) is int, "All keys in overlap mapping must be integer."
                assert type(v) is list, "All values in overlap mapping must be list."
                for i in v:
                    assert type(i) is int, \
                        "All elements in the overlap mapping values must be integers."
            self._overlapping = overlapping
        else:
            self._overlapping = {}

    @property
    def rows(self):
        """
        The number of rows in the grid.
        """
        return self._internal.shape[0]

    @property
    def cols(self):
        """
        The number of columns in the grid.
        """
        return self._internal.shape[1]

    def reset(self, **kwargs):
        """
        Reset the grid to an empty state.
        """
        for i in range(self.rows):
            for j in range(self.cols):
                self._internal[i, j] = {}

    def query(self, agent, ndx):
        """
        Query a cell in the grid to see if is available to this agent.

        The cell is available for the agent if it is empty or if both the occupying agent
        and the querying agent are overlappable.

        Args:
            agent: The agent for which we are checking availabilty.
            ndx: The cell to query.

        Returns:
            The availability of this cell.
        """
        ndx = tuple(ndx)
        if self._internal[ndx]: # There are agents here
            try:
                return all([
                    True if other.encoding in self._overlapping[agent.encoding] else False
                    for other in self._internal[ndx].values() if other.active
                ])
            except KeyError:
                return False
        else:
            return True

    def place(self, agent, ndx):
        """
        Place an agent at an index.

        If the cell is available, the agent will be placed at that index
        in the grid and the agent's position will be updated. The placement is
        successful if the new position is unoccupied or if the agent already occupying
        that position is overlappable AND this agent is overlappable.

        Args:
            agent: The agent to place.
            ndx: The new index for this agent.

        Returns:
            The successfulness of the placement.
        """
        ndx = tuple(ndx)
        if self.query(agent, ndx):
            self._internal[ndx][agent.id] = agent
            agent.position = np.array(ndx)
            return True
        else:
            return False

    def remove(self, agent, ndx):
        """
        Remove an agent from an index.

        Args:
            agent: The agent to remove
            ndx: The old index for this agent
        """
        ndx = tuple(ndx)
        del self._internal[ndx][agent.id]

    def __getitem__(self, subscript):
        return self._internal.__getitem__(subscript)



import numpy as np
import pytest

from abmarl.sim.corridor import MultiCorridor as Corridor
from abmarl.managers import TurnBasedManager


def test_init():
    sim = Corridor()
    wrapped_sim = TurnBasedManager(sim)
    assert wrapped_sim.sim == sim
    assert wrapped_sim.agents == sim.agents
    assert next(wrapped_sim.agent_order) == 'agent0'
    assert next(wrapped_sim.agent_order) == 'agent1'
    assert next(wrapped_sim.agent_order) == 'agent2'
    assert next(wrapped_sim.agent_order) == 'agent3'
    assert next(wrapped_sim.agent_order) == 'agent4'


def test_reset_and_step():
    np.random.seed(24)
    sim = TurnBasedManager(Corridor())

    obs = sim.reset()
    assert sim.sim.corridor[4].id == 'agent3'
    assert sim.sim.corridor[5].id == 'agent4'
    assert sim.sim.corridor[6].id == 'agent2'
    assert sim.sim.corridor[7].id == 'agent1'
    assert sim.sim.corridor[8].id == 'agent0'
    assert sim.done_agents == set()
    assert obs == {'agent0': {'left': [True], 'position': [8], 'right': [False]}}

    obs, reward, done, info = sim.step({agent_id: Corridor.Actions.RIGHT for agent_id in obs})
    assert obs == {'agent1': {'left': [True], 'position': [7], 'right': [False]}}
    assert reward == {'agent1': 0}
    assert done == {'agent1': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: Corridor.Actions.RIGHT for agent_id in obs})
    assert obs == {'agent2': {'left': [True], 'position': [6], 'right': [False]}}
    assert reward == {'agent2': 0}
    assert done == {'agent2': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: Corridor.Actions.RIGHT for agent_id in obs})
    assert obs == {'agent3': {'left': [False], 'position': [4], 'right': [True]}}
    assert reward == {'agent3': 0}
    assert done == {'agent3': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: Corridor.Actions.RIGHT for agent_id in obs})
    assert obs == {'agent4': {'left': [True], 'position': [5], 'right': [False]}}
    assert reward == {'agent4': -2}
    assert done == {'agent4': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: Corridor.Actions.RIGHT for agent_id in obs})
    assert obs == {
        'agent0': {'left': [True], 'position': [9], 'right': [False]},
        'agent1': {'left': [True], 'position': [8], 'right': [False]}}
    assert reward == {'agent0': 100, 'agent1': -1}
    assert done == {'agent0': True, 'agent1': False, '__all__': False}

    with pytest.raises(AssertionError):
        sim.step({'agent0': Corridor.Actions.STAY})

    obs, reward, done, info = sim.step({'agent1': Corridor.Actions.STAY})
    assert obs == {'agent2': {'left': [True], 'position': [7], 'right': [True]}}
    assert reward == {'agent2': -1,}
    assert done == {'agent2': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: Corridor.Actions.LEFT for agent_id in obs})
    assert obs == {'agent3': {'left': [False], 'position': [4], 'right': [False]}}
    assert reward == {'agent3': -5}
    assert done == {'agent3': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: Corridor.Actions.STAY for agent_id in obs})
    assert obs == {'agent4': {'left': [False], 'position': [6], 'right': [True]}}
    assert reward == {'agent4': -3}
    assert done == {'agent4': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: Corridor.Actions.LEFT for agent_id in obs})
    assert obs == {'agent1': {'left': [True], 'position': [8], 'right': [False]}}
    assert reward == {'agent1': -1}
    assert done == {'agent1': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: Corridor.Actions.RIGHT for agent_id in obs})
    assert obs == {'agent2': {'left': [False], 'position': [7], 'right': [False]}}
    assert reward == {'agent2': -5}
    assert done == {'agent2': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: Corridor.Actions.RIGHT for agent_id in obs})
    assert obs == {'agent3': {'left': [False], 'position': [4], 'right': [True]}}
    assert reward == {'agent3': -1}
    assert done == {'agent3': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: Corridor.Actions.RIGHT for agent_id in obs})
    assert obs == {'agent4': {'left': [True], 'position': [5], 'right': [False]}}
    assert reward == {'agent4': -3}
    assert done == {'agent4': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: Corridor.Actions.LEFT for agent_id in obs})
    assert obs == {
        'agent1': {'left': [True], 'position': [9], 'right': [False]},
        'agent2': {'left': [False], 'position': [8], 'right': [False]}}
    assert reward == {'agent1': 100, 'agent2': -1}
    assert done == {'agent1': True, 'agent2': False, '__all__': False}

    with pytest.raises(AssertionError):
        sim.step({'agent1': Corridor.Actions.STAY})

    obs, reward, done, info = sim.step({'agent2': Corridor.Actions.STAY})
    assert obs == {'agent3': {'left': [False], 'position': [4], 'right': [True]}}
    assert reward == {'agent3': -7,}
    assert done == {'agent3': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: Corridor.Actions.LEFT for agent_id in obs})
    assert obs == {'agent4': {'left': [False], 'position': [5], 'right': [False]}}
    assert reward == {'agent4': -5,}
    assert done == {'agent4': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: Corridor.Actions.RIGHT for agent_id in obs})
    assert obs == {'agent2': {'left': [False], 'position': [8], 'right': [False]}}
    assert reward == {'agent2': -1,}
    assert done == {'agent2': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: Corridor.Actions.RIGHT for agent_id in obs})
    assert obs == {'agent3': {'left': [False], 'position': [3], 'right': [False]}}
    assert reward == {'agent3': -1,}
    assert done == {'agent3': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: Corridor.Actions.RIGHT for agent_id in obs})
    assert obs == {'agent4': {'left': [False], 'position': [6], 'right': [False]}}
    assert reward == {'agent4': -1,}
    assert done == {'agent4': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: Corridor.Actions.RIGHT for agent_id in obs})
    assert obs == {
        'agent2': {'left': [False], 'position': [9], 'right': [False]},
        'agent3': {'left': [False], 'position': [4], 'right': [False]}}
    assert reward == {'agent2': 100, 'agent3': -1}
    assert done == {'agent2': True, 'agent3': False, '__all__': False}

    with pytest.raises(AssertionError):
        sim.step({'agent2': Corridor.Actions.STAY})

    obs, reward, done, info = sim.step({'agent3': Corridor.Actions.RIGHT})
    assert obs == {'agent4': {'left': [False], 'position': [7], 'right': [False]}}
    assert reward == {'agent4': -1,}
    assert done == {'agent4': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: Corridor.Actions.RIGHT for agent_id in obs})
    assert obs == {'agent3': {'left': [False], 'position': [5], 'right': [False]}}
    assert reward == {'agent3': -1,}
    assert done == {'agent3': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: Corridor.Actions.RIGHT for agent_id in obs})
    assert obs == {'agent4': {'left': [False], 'position': [8], 'right': [False]}}
    assert reward == {'agent4': -1,}
    assert done == {'agent4': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: Corridor.Actions.RIGHT for agent_id in obs})
    assert obs == {'agent3': {'left': [False], 'position': [6], 'right': [False]}}
    assert reward == {'agent3': -1,}
    assert done == {'agent3': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: Corridor.Actions.RIGHT for agent_id in obs})
    assert obs == {
        'agent4': {'left': [False], 'position': [9], 'right': [False]},
        'agent3': {'left': [False], 'position': [7], 'right': [False]}}
    assert reward == {'agent4': 100, 'agent3': -1}
    assert done == {'agent4': True, 'agent3': False, '__all__': False}

    with pytest.raises(AssertionError):
        sim.step({'agent4': Corridor.Actions.STAY})

    obs, reward, done, info = sim.step({'agent3': Corridor.Actions.RIGHT})
    assert obs == {'agent3': {'left': [False], 'position': [8], 'right': [False]}}
    assert reward == {'agent3': -1,}
    assert done == {'agent3': False, '__all__': False}

    obs, reward, done, info = sim.step({agent_id: Corridor.Actions.RIGHT for agent_id in obs})
    assert obs == {'agent3': {'left': [False], 'position': [9], 'right': [False]}}
    assert reward == {'agent3': 100,}
    assert done == {'agent3': True, '__all__': True}




from gym.spaces import Discrete, Box
import numpy as np

from abmarl.sim.gridworld.actor import MoveActor, ActorBaseComponent
from abmarl.sim.gridworld.state import PositionState
from abmarl.sim.gridworld.wrapper import RavelActionWrapper, ActorWrapper

from .helpers import grid, moving_agents as agents


position_state = PositionState(grid=grid, agents=agents)
move_actor = MoveActor(grid=grid, agents=agents)
ravel_action_wrapper = RavelActionWrapper(move_actor)


def test_ravel_action_wrapper_properties():
    assert isinstance(ravel_action_wrapper, ActorWrapper)
    assert isinstance(ravel_action_wrapper, ActorBaseComponent)
    assert ravel_action_wrapper.wrapped_component == move_actor
    assert ravel_action_wrapper.unwrapped == move_actor
    assert ravel_action_wrapper.agents == move_actor.agents
    assert ravel_action_wrapper.grid == move_actor.grid
    assert ravel_action_wrapper.key == move_actor.key
    assert ravel_action_wrapper.supported_agent_type == move_actor.supported_agent_type


def test_ravel_action_wrapper_agent_spaces():
    assert ravel_action_wrapper.from_space['agent0'] == Box(-1, 1, (2,), int)
    assert ravel_action_wrapper.agents['agent0'].action_space['move'] == Discrete(9)
    assert ravel_action_wrapper.from_space['agent1'] == Box(-2, 2, (2,), int)
    assert ravel_action_wrapper.agents['agent1'].action_space['move'] == Discrete(25)
    assert ravel_action_wrapper.from_space['agent2'] == Box(-1, 1, (2,), int)
    assert ravel_action_wrapper.agents['agent2'].action_space['move'] == Discrete(9)
    assert ravel_action_wrapper.from_space['agent3'] == Box(-3, 3, (2,), int)
    assert ravel_action_wrapper.agents['agent3'].action_space['move'] == Discrete(49)


def test_ravel_action_wrapper_process_action():
    action_sample = {
        'agent0': {'move': 7},
        'agent1': {'move': 3},
        'agent2': {'move': 4},
        'agent3': {'move': 34},
    }
    np.testing.assert_array_equal(
        ravel_action_wrapper.wrap_point(Box(-1, 1, (2,), int), 7),
        np.array([1, 0])
    )
    np.testing.assert_array_equal(
        ravel_action_wrapper.wrap_point(Box(-2, 2, (2,), int), 3),
        np.array([-2, 1])
    )
    np.testing.assert_array_equal(
        ravel_action_wrapper.wrap_point(Box(-1, 1, (2,), int), 4),
        np.array([0, 0])
    )
    np.testing.assert_array_equal(
        ravel_action_wrapper.wrap_point(Box(-3, 3, (2,), int), 34),
        np.array([1, 3])
    )

    position_state.reset()
    assert ravel_action_wrapper.process_action(agents['agent0'], action_sample['agent0'])
    np.testing.assert_array_equal(agents['agent0'].position, np.array([4, 4]))
    assert ravel_action_wrapper.process_action(agents['agent1'], action_sample['agent1'])
    np.testing.assert_array_equal(agents['agent1'].position, np.array([0, 3]))
    assert ravel_action_wrapper.process_action(agents['agent2'], action_sample['agent2'])
    np.testing.assert_array_equal(agents['agent2'].position, np.array([0, 1]))
    assert not ravel_action_wrapper.process_action(agents['agent3'], action_sample['agent3'])
    np.testing.assert_array_equal(agents['agent3'].position, np.array([3, 1]))



from .wrapper import Wrapper

# SAR-based wrappers
from .sar_wrapper import SARWrapper
from .flatten_wrapper import FlattenActionWrapper, FlattenWrapper
from .ravel_discrete_wrapper import RavelDiscreteWrapper
from .super_agent_wrapper import SuperAgentWrapper

from .communication_wrapper import CommunicationHandshakeWrapper




import numpy as np
import pytest

from abmarl.sim.gridworld.grid import Grid
from abmarl.sim.gridworld.agent import GridWorldAgent


def test_grid_rows():
    grid = Grid(2, 6)
    assert grid.rows == 2

    with pytest.raises(AssertionError):
        Grid(-1, 3)


def test_grid_cols():
    grid = Grid(2, 6)
    assert grid.cols == 6

    with pytest.raises(AssertionError):
        Grid(1, 5.0)


def test_grid_internal():
    grid = Grid(2, 6)
    np.testing.assert_equal(grid._internal, np.empty((2, 6), dtype=object))


def test_grid_overlapping():
    grid = Grid(3, 3, overlapping={1: [2], 2: [1], 3: [1, 2]})
    assert grid._overlapping == {1: [2], 2: [1], 3: [1, 2]}

    with pytest.raises(AssertionError):
        Grid(2, 2, overlapping=[1, 2, 3])

    with pytest.raises(AssertionError):
        Grid(2, 2, overlapping={'1': [3], 2.0: [6]})

    with pytest.raises(AssertionError):
        Grid(2, 2, overlapping={1: 3, 2: [6]})

    with pytest.raises(AssertionError):
        Grid(2, 2, overlapping={1: ['2', 3], 2: [2, 3]})


def test_grid_reset():
    grid = Grid(3, 3)
    grid.reset()
    for i in range(grid.rows):
        for j in range(grid.cols):
            assert grid[i, j] == {}

    agent = GridWorldAgent(id='agent0', encoding=1)
    assert grid.place(agent, (1, 0))
    assert grid[1, 0] == {'agent0': agent}
    for i in range(grid.rows):
        for j in range(grid.cols):
            if i == 1 and j == 0:
                continue
            assert grid[i, j] == {}
    grid.reset()
    assert grid[1, 0] == {}


def test_grid_query():
    grid = Grid(3, 3)
    grid.reset()
    agent1 = GridWorldAgent(id='agent1', encoding=1)
    agent2 = GridWorldAgent(id='agent2', encoding=1)
    assert grid.place(agent1, (1, 0))
    assert not grid.query(agent2, (1, 0))
    assert grid.query(agent2, (0, 1))


def test_grid_query_with_overlap():
    grid = Grid(3, 3, overlapping={1: [1]})
    grid.reset()
    agent1 = GridWorldAgent(id='agent1', encoding=1)
    agent2 = GridWorldAgent(id='agent2', encoding=1)
    assert grid.place(agent1, (1, 0))
    assert grid.query(agent2, (1, 0))
    assert grid.query(agent1, (1, 0))


def test_grid_query_overlap_with_inactive_agent():
    grid = Grid(3, 3)
    grid.reset()
    agent1 = GridWorldAgent(id='agent1', encoding=1)
    agent2 = GridWorldAgent(id='agent2', encoding=1)
    assert grid.place(agent1, (1, 0))
    agent1._active = False
    assert grid.query(agent2, (1, 0))


def test_grid_place():
    grid = Grid(3, 3)
    grid.reset()
    agent1 = GridWorldAgent(id='agent1', encoding=1)
    agent2 = GridWorldAgent(id='agent2', encoding=2)
    assert grid.place(agent1, (1, 0))
    assert not grid.place(agent2, (1, 0))


def test_grid_place_with_overlap():
    grid = Grid(3, 3, overlapping={1: [2], 2: [1]})
    grid.reset()
    agent1 = GridWorldAgent(id='agent1', encoding=1)
    agent2 = GridWorldAgent(id='agent2', encoding=2)
    assert grid.place(agent1, (1, 0))
    assert grid.place(agent2, (1, 0))
    assert grid[1, 0] == {'agent1': agent1, 'agent2': agent2}


def test_grid_remove():
    grid = Grid(3, 3, overlapping={1: [2], 2: [1]})
    grid.reset()
    agent1 = GridWorldAgent(id='agent1', encoding=1)
    agent2 = GridWorldAgent(id='agent2', encoding=2)
    agent3 = GridWorldAgent(id='agent3', encoding=2)
    assert grid.place(agent1, (1, 0))
    assert grid.place(agent2, (1, 0))
    assert grid.place(agent3, (0, 1))
    assert grid[1, 0] == {'agent1': agent1, 'agent2': agent2}
    assert grid[0, 1] == {'agent3': agent3}

    grid.remove(agent1, (1, 0))
    grid.remove(agent3, (0, 1))
    assert grid[1, 0] == {'agent2': agent2}
    assert grid[0, 1] == {}

    with pytest.raises(KeyError):
        grid.remove(agent1, (1, 0))



import numpy as np
import pytest

from abmarl.sim.corridor import MultiCorridor as Corridor
from abmarl.managers import AllStepManager


def test_init():
    sim = Corridor()
    wrapped_sim = AllStepManager(sim)
    assert wrapped_sim.sim == sim
    assert wrapped_sim.agents == sim.agents


def test_reset_and_step():
    np.random.seed(24)
    sim = AllStepManager(Corridor())

    obs = sim.reset()
    assert sim.sim.corridor[4].id == 'agent3'
    assert sim.sim.corridor[5].id == 'agent4'
    assert sim.sim.corridor[6].id == 'agent2'
    assert sim.sim.corridor[7].id == 'agent1'
    assert sim.sim.corridor[8].id == 'agent0'
    assert sim.done_agents == set()
    assert obs['agent0'] == {'left': [True], 'position': [8], 'right': [False]}
    assert obs['agent1'] == {'left': [True], 'position': [7], 'right': [True]}
    assert obs['agent2'] == {'left': [True], 'position': [6], 'right': [True]}
    assert obs['agent3'] == {'left': [False], 'position': [4], 'right': [True]}
    assert obs['agent4'] == {'left': [True], 'position': [5], 'right': [True]}


    obs, reward, done, _ = sim.step({
        'agent0': Corridor.Actions.RIGHT,
        'agent1': Corridor.Actions.RIGHT,
        'agent2': Corridor.Actions.RIGHT,
        'agent3': Corridor.Actions.RIGHT,
        'agent4': Corridor.Actions.RIGHT,
    })

    assert obs['agent0'] == {'left': [True], 'position': [9], 'right': [False]}
    assert obs['agent1'] == {'left': [True], 'position': [8], 'right': [False]}
    assert obs['agent2'] == {'left': [True], 'position': [7], 'right': [True]}
    assert obs['agent3'] == {'left': [False], 'position': [4], 'right': [False]}
    assert obs['agent4'] == {'left': [False], 'position': [6], 'right': [True]}
    assert reward['agent0'] == 100
    assert reward['agent1'] == -1
    assert reward['agent2'] == -1
    assert reward['agent3'] == -5
    assert reward['agent4'] == -3
    assert done['agent0']
    assert not done['agent1']
    assert not done['agent2']
    assert not done['agent3']
    assert not done['agent4']
    assert not done['__all__']


    with pytest.raises(AssertionError):
        sim.step({
            'agent0': Corridor.Actions.RIGHT,
            'agent1': Corridor.Actions.STAY,
            'agent2': Corridor.Actions.LEFT,
            'agent3': Corridor.Actions.STAY,
            'agent4': Corridor.Actions.LEFT,
        })

    obs, reward, done, _ = sim.step({
        'agent1': Corridor.Actions.STAY,
        'agent2': Corridor.Actions.LEFT,
        'agent3': Corridor.Actions.STAY,
        'agent4': Corridor.Actions.LEFT,
    })

    assert 'agent0' not in obs
    assert obs['agent1'] == {'left': [True], 'position': [8], 'right': [False]}
    assert obs['agent2'] == {'left': [False], 'position': [7], 'right': [True]}
    assert obs['agent3'] == {'left': [False], 'position': [4], 'right': [True]}
    assert obs['agent4'] == {'left': [True], 'position': [5], 'right': [False]}
    assert 'agent0' not in reward
    assert reward['agent1'] == -1
    assert reward['agent2'] == -5
    assert reward['agent3'] == -1
    assert reward['agent4'] == -3
    assert 'agent0' not in done
    assert not done['agent1']
    assert not done['agent2']
    assert not done['agent3']
    assert not done['agent4']
    assert not done['__all__']


    obs, reward, done, _ = sim.step({
        'agent1': Corridor.Actions.RIGHT,
        'agent2': Corridor.Actions.RIGHT,
        'agent3': Corridor.Actions.RIGHT,
        'agent4': Corridor.Actions.LEFT,
    })

    assert obs['agent1'] == {'left': [True], 'position': [9], 'right': [False]}
    assert obs['agent2'] == {'left': [False], 'position': [8], 'right': [False]}
    assert obs['agent3'] == {'left': [False], 'position': [4], 'right': [True]}
    assert obs['agent4'] == {'left': [True], 'position': [5], 'right': [False]}
    assert reward['agent1'] == 100
    assert reward['agent2'] == -1
    assert reward['agent3'] == -7
    assert reward['agent4'] == -7
    assert done['agent1']
    assert not done['agent2']
    assert not done['agent3']
    assert not done['agent4']
    assert not done['__all__']


    with pytest.raises(AssertionError):
        sim.step({
            'agent1': Corridor.Actions.STAY,
            'agent2': Corridor.Actions.STAY,
            'agent3': Corridor.Actions.LEFT,
            'agent4': Corridor.Actions.RIGHT,
        })

    obs, reward, done, _ = sim.step({
        'agent2': Corridor.Actions.STAY,
        'agent3': Corridor.Actions.LEFT,
        'agent4': Corridor.Actions.RIGHT,
    })

    assert 'agent1' not in obs
    assert obs['agent2'] == {'left': [False], 'position': [8], 'right': [False]}
    assert obs['agent3'] == {'left': [False], 'position': [3], 'right': [False]}
    assert obs['agent4'] == {'left': [False], 'position': [6], 'right': [False]}
    assert 'agent1' not in reward
    assert reward['agent2'] == -1
    assert reward['agent3'] == -1
    assert reward['agent4'] == -1
    assert 'agent1' not in done
    assert not done['agent2']
    assert not done['agent3']
    assert not done['agent4']
    assert not done['__all__']


    obs, reward, done, _ = sim.step({
        'agent2': Corridor.Actions.RIGHT,
        'agent3': Corridor.Actions.RIGHT,
        'agent4': Corridor.Actions.RIGHT,
    })

    assert obs['agent2'] == {'left': [False], 'position': [9], 'right': [False]}
    assert obs['agent3'] == {'left': [False], 'position': [4], 'right': [False]}
    assert obs['agent4'] == {'left': [False], 'position': [7], 'right': [False]}
    assert reward['agent2'] == 100
    assert reward['agent3'] == -1
    assert reward['agent4'] == -1
    assert done['agent2']
    assert not done['agent3']
    assert not done['agent4']
    assert not done['__all__']


    with pytest.raises(AssertionError):
        sim.step({
            'agent2': Corridor.Actions.STAY,
            'agent3': Corridor.Actions.RIGHT,
            'agent4': Corridor.Actions.RIGHT,
        })

    obs, reward, done, _ = sim.step({
        'agent3': Corridor.Actions.RIGHT,
        'agent4': Corridor.Actions.RIGHT,
    })

    assert 'agent2' not in obs
    assert obs['agent3'] == {'left': [False], 'position': [5], 'right': [False]}
    assert obs['agent4'] == {'left': [False], 'position': [8], 'right': [False]}
    assert 'agent2' not in reward
    assert reward['agent3'] == -1
    assert reward['agent4'] == -1
    assert 'agent2' not in done
    assert not done['agent3']
    assert not done['agent4']
    assert not done['__all__']


    obs, reward, done, _ = sim.step({
        'agent3': Corridor.Actions.RIGHT,
        'agent4': Corridor.Actions.RIGHT,
    })

    assert obs['agent3'] == {'left': [False], 'position': [6], 'right': [False]}
    assert obs['agent4'] == {'left': [False], 'position': [9], 'right': [False]}
    assert reward['agent3'] == -1
    assert reward['agent4'] == 100
    assert not done['agent3']
    assert done['agent4']
    assert not done['__all__']


    with pytest.raises(AssertionError):
        sim.step({
            'agent3': Corridor.Actions.RIGHT,
            'agent4': Corridor.Actions.STAY,
        })

    obs, reward, done, _ = sim.step({
        'agent3': Corridor.Actions.RIGHT,
    })

    assert 'agent4' not in obs
    assert obs == {'agent3': {'left': [False], 'position': [7], 'right': [False]}}
    assert 'agent4' not in reward
    assert reward == {'agent3': -1,}
    assert 'agent4' not in done
    assert done == {'agent3': False, '__all__': False}


    obs, reward, done, _ = sim.step({
        'agent3': Corridor.Actions.RIGHT,
    })

    assert obs == {'agent3': {'left': [False], 'position': [8], 'right': [False]}}
    assert reward == {'agent3': -1,}
    assert done == {'agent3': False, '__all__': False}


    obs, reward, done, _ = sim.step({
        'agent3': Corridor.Actions.RIGHT,
    })

    assert obs == {'agent3': {'left': [False], 'position': [9], 'right': [False]}}
    assert reward == {'agent3': 100}
    assert done == {'agent3': True, '__all__': True}



from abmarl.sim.wrappers import SARWrapper
from .helpers import MultiAgentSim


# Individual wrappers
class ObservationWrapper(SARWrapper):
    def wrap_observation(self, from_agent, observation):
        return "Wrap Observation: " + observation

    def unwrap_observation(self, from_agent, observation):
        return "Unwrap Observation: " + observation


class ActionWrapper(SARWrapper):
    def wrap_action(self, from_agent, action):
        return "Wrap Action: " + str(action)

    def unwrap_action(self, from_agent, action):
        return "Unwrap Action: " + str(action)


class RewardWrapper(SARWrapper):
    def wrap_reward(self, reward):
        return "Wrap Reward: " + reward

    def unwrap_reward(self, reward):
        return "Unwrap Reward: " + reward


# Combination wrappers
class ObservationActionWrapper(ObservationWrapper, ActionWrapper):
    pass


class ObservationRewardWrapper(ObservationWrapper, RewardWrapper):
    pass


class ActionRewardWrapper(ActionWrapper, RewardWrapper):
    pass


class ObservationActionRewardWrapper(ObservationWrapper, ActionWrapper, RewardWrapper):
    pass


# Tests
def test_sar_wrapped():
    sim = MultiAgentSim(5)
    wrapped_sim = SARWrapper(sim)
    assert wrapped_sim.sim == sim
    assert wrapped_sim.unwrapped == sim
    assert wrapped_sim.agents == sim.agents


def test_wrap_observation():
    sim = ObservationWrapper(MultiAgentSim())
    sim.reset()
    assert sim.get_obs('agent0') == 'Wrap Observation: Obs from agent0'
    assert sim.get_obs('agent1') == 'Wrap Observation: Obs from agent1'
    assert sim.get_obs('agent2') == 'Wrap Observation: Obs from agent2'

    sim.step({agent_id: i for i, agent_id in enumerate(sim.agents)})
    assert sim.get_obs('agent0') == 'Wrap Observation: Obs from agent0'
    assert sim.get_obs('agent1') == 'Wrap Observation: Obs from agent1'
    assert sim.get_obs('agent2') == 'Wrap Observation: Obs from agent2'
    assert sim.get_reward('agent0') == 'Reward from agent0'
    assert sim.get_reward('agent1') == 'Reward from agent1'
    assert sim.get_reward('agent2') == 'Reward from agent2'
    assert sim.get_done('agent0') == 'Done from agent0'
    assert sim.get_done('agent1') == 'Done from agent1'
    assert sim.get_done('agent2') == 'Done from agent2'
    assert sim.get_all_done() == "Done from all agents and/or simulation."
    assert sim.get_info('agent0') == {'Action from agent0': 0}
    assert sim.get_info('agent1') == {'Action from agent1': 1}
    assert sim.get_info('agent2') == {'Action from agent2': 2}


def test_wrap_action():
    sim = ActionWrapper(MultiAgentSim())
    sim.reset()
    assert sim.get_obs('agent0') == 'Obs from agent0'
    assert sim.get_obs('agent1') == 'Obs from agent1'
    assert sim.get_obs('agent2') == 'Obs from agent2'

    sim.step({agent_id: i for i, agent_id in enumerate(sim.agents)})
    assert sim.get_obs('agent0') == 'Obs from agent0'
    assert sim.get_obs('agent1') == 'Obs from agent1'
    assert sim.get_obs('agent2') == 'Obs from agent2'
    assert sim.get_reward('agent0') == 'Reward from agent0'
    assert sim.get_reward('agent1') == 'Reward from agent1'
    assert sim.get_reward('agent2') == 'Reward from agent2'
    assert sim.get_done('agent0') == 'Done from agent0'
    assert sim.get_done('agent1') == 'Done from agent1'
    assert sim.get_done('agent2') == 'Done from agent2'
    assert sim.get_all_done() == "Done from all agents and/or simulation."
    assert sim.get_info('agent0') == {'Action from agent0': 'Wrap Action: 0'}
    assert sim.get_info('agent1') == {'Action from agent1': 'Wrap Action: 1'}
    assert sim.get_info('agent2') == {'Action from agent2': 'Wrap Action: 2'}


def test_wrap_reward():
    sim = RewardWrapper(MultiAgentSim())
    sim.reset()
    assert sim.get_obs('agent0') == 'Obs from agent0'
    assert sim.get_obs('agent1') == 'Obs from agent1'
    assert sim.get_obs('agent2') == 'Obs from agent2'

    sim.step({agent_id: i for i, agent_id in enumerate(sim.agents)})
    assert sim.get_obs('agent0') == 'Obs from agent0'
    assert sim.get_obs('agent1') == 'Obs from agent1'
    assert sim.get_obs('agent2') == 'Obs from agent2'
    assert sim.get_reward('agent0') == 'Wrap Reward: Reward from agent0'
    assert sim.get_reward('agent1') == 'Wrap Reward: Reward from agent1'
    assert sim.get_reward('agent2') == 'Wrap Reward: Reward from agent2'
    assert sim.get_done('agent0') == 'Done from agent0'
    assert sim.get_done('agent1') == 'Done from agent1'
    assert sim.get_done('agent2') == 'Done from agent2'
    assert sim.get_all_done() == "Done from all agents and/or simulation."
    assert sim.get_info('agent0') == {'Action from agent0': 0}
    assert sim.get_info('agent1') == {'Action from agent1': 1}
    assert sim.get_info('agent2') == {'Action from agent2': 2}


def test_wrap_observation_action():
    sim = ObservationActionWrapper(MultiAgentSim())
    sim.reset()
    assert sim.get_obs('agent0') == 'Wrap Observation: Obs from agent0'
    assert sim.get_obs('agent1') == 'Wrap Observation: Obs from agent1'
    assert sim.get_obs('agent2') == 'Wrap Observation: Obs from agent2'

    sim.step({agent_id: i for i, agent_id in enumerate(sim.agents)})
    assert sim.get_obs('agent0') == 'Wrap Observation: Obs from agent0'
    assert sim.get_obs('agent1') == 'Wrap Observation: Obs from agent1'
    assert sim.get_obs('agent2') == 'Wrap Observation: Obs from agent2'
    assert sim.get_reward('agent0') == 'Reward from agent0'
    assert sim.get_reward('agent1') == 'Reward from agent1'
    assert sim.get_reward('agent2') == 'Reward from agent2'
    assert sim.get_done('agent0') == 'Done from agent0'
    assert sim.get_done('agent1') == 'Done from agent1'
    assert sim.get_done('agent2') == 'Done from agent2'
    assert sim.get_all_done() == "Done from all agents and/or simulation."
    assert sim.get_info('agent0') == {'Action from agent0': 'Wrap Action: 0'}
    assert sim.get_info('agent1') == {'Action from agent1': 'Wrap Action: 1'}
    assert sim.get_info('agent2') == {'Action from agent2': 'Wrap Action: 2'}


def test_wrap_observation_reward():
    sim = ObservationRewardWrapper(MultiAgentSim())
    sim.reset()
    assert sim.get_obs('agent0') == 'Wrap Observation: Obs from agent0'
    assert sim.get_obs('agent1') == 'Wrap Observation: Obs from agent1'
    assert sim.get_obs('agent2') == 'Wrap Observation: Obs from agent2'

    sim.step({agent_id: i for i, agent_id in enumerate(sim.agents)})
    assert sim.get_obs('agent0') == 'Wrap Observation: Obs from agent0'
    assert sim.get_obs('agent1') == 'Wrap Observation: Obs from agent1'
    assert sim.get_obs('agent2') == 'Wrap Observation: Obs from agent2'
    assert sim.get_reward('agent0') == 'Wrap Reward: Reward from agent0'
    assert sim.get_reward('agent1') == 'Wrap Reward: Reward from agent1'
    assert sim.get_reward('agent2') == 'Wrap Reward: Reward from agent2'
    assert sim.get_done('agent0') == 'Done from agent0'
    assert sim.get_done('agent1') == 'Done from agent1'
    assert sim.get_done('agent2') == 'Done from agent2'
    assert sim.get_info('agent0') == {'Action from agent0': 0}
    assert sim.get_info('agent1') == {'Action from agent1': 1}
    assert sim.get_info('agent2') == {'Action from agent2': 2}


def test_wrap_action_reward():
    sim = ActionRewardWrapper(MultiAgentSim())
    sim.reset()
    assert sim.get_obs('agent0') == 'Obs from agent0'
    assert sim.get_obs('agent1') == 'Obs from agent1'
    assert sim.get_obs('agent2') == 'Obs from agent2'

    sim.step({agent_id: i for i, agent_id in enumerate(sim.agents)})
    assert sim.get_obs('agent0') == 'Obs from agent0'
    assert sim.get_obs('agent1') == 'Obs from agent1'
    assert sim.get_obs('agent2') == 'Obs from agent2'
    assert sim.get_reward('agent0') == 'Wrap Reward: Reward from agent0'
    assert sim.get_reward('agent1') == 'Wrap Reward: Reward from agent1'
    assert sim.get_reward('agent2') == 'Wrap Reward: Reward from agent2'
    assert sim.get_done('agent0') == 'Done from agent0'
    assert sim.get_done('agent1') == 'Done from agent1'
    assert sim.get_done('agent2') == 'Done from agent2'
    assert sim.get_info('agent0') == {'Action from agent0': 'Wrap Action: 0'}
    assert sim.get_info('agent1') == {'Action from agent1': 'Wrap Action: 1'}
    assert sim.get_info('agent2') == {'Action from agent2': 'Wrap Action: 2'}


def test_wrap_observation_action_reward():
    sim = ObservationActionRewardWrapper(MultiAgentSim())
    sim.reset()
    assert sim.get_obs('agent0') == 'Wrap Observation: Obs from agent0'
    assert sim.get_obs('agent1') == 'Wrap Observation: Obs from agent1'
    assert sim.get_obs('agent2') == 'Wrap Observation: Obs from agent2'

    sim.step({agent_id: i for i, agent_id in enumerate(sim.agents)})
    assert sim.get_obs('agent0') == 'Wrap Observation: Obs from agent0'
    assert sim.get_obs('agent1') == 'Wrap Observation: Obs from agent1'
    assert sim.get_obs('agent2') == 'Wrap Observation: Obs from agent2'
    assert sim.get_reward('agent0') == 'Wrap Reward: Reward from agent0'
    assert sim.get_reward('agent1') == 'Wrap Reward: Reward from agent1'
    assert sim.get_reward('agent2') == 'Wrap Reward: Reward from agent2'
    assert sim.get_done('agent0') == 'Done from agent0'
    assert sim.get_done('agent1') == 'Done from agent1'
    assert sim.get_done('agent2') == 'Done from agent2'
    assert sim.get_info('agent0') == {'Action from agent0': 'Wrap Action: 0'}
    assert sim.get_info('agent1') == {'Action from agent1': 'Wrap Action: 1'}
    assert sim.get_info('agent2') == {'Action from agent2': 'Wrap Action: 2'}




from abc import abstractmethod

from abmarl.sim.gridworld.actor import ActorBaseComponent
from abmarl.sim.gridworld.observer import ObserverBaseComponent
from abmarl.sim.gridworld.base import GridWorldBaseComponent
from abmarl.sim.wrappers import ravel_discrete_wrapper as rdw


class ComponentWrapper(GridWorldBaseComponent):
    """
    Wraps GridWorldBaseComponent.

    Every wrapper must be able to wrap the respective space and points
    to/from that space. Agents and Grid are referenced directly from the wrapped
    component rather than received as initialization parameters.
    """
    @property
    @abstractmethod
    def wrapped_component(self):
        """
        Get the first-level wrapped component.
        """
        pass

    @property
    def unwrapped(self):
        """
        Fall through all the wrappers and obtain the original, completely unwrapped component.
        """
        try:
            return self.wrapped_component.unwrapped
        except AttributeError:
            return self.wrapped_component

    @property
    def agents(self):
        """
        The agent dictionary is directly taken from the wrapped component.
        """
        return self.wrapped_component.agents

    @property
    def grid(self):
        """
        The grid is directly taken from the wrapped component.
        """
        return self.wrapped_component.grid

    @abstractmethod
    def check_space(self, space):
        """
        Verify that the space can be wrapped.
        """
        pass

    @abstractmethod
    def wrap_space(self, space):
        """
        Wrap the space.

        Args:
            space: The space to wrap.
        """
        pass

    @abstractmethod
    def wrap_point(self, space, point):
        """
        Wrap a point to the space.

        Args:
            space: The space into which to wrap the point.
            point: The point to wrap.
        """
        pass


class ActorWrapper(ComponentWrapper, ActorBaseComponent):
    """
    Wraps an ActorComponent.

    Modify the action space of the agents involved with the Actor, namely the specific
    actor's channel. The actions recieved from the trainer are in the wrapped space,
    so we need to unwrap them to send them to the actor. This is the opposite from
    how we wrap and unwrap observations.
    """
    def __init__(self, component):
        assert isinstance(component, ActorBaseComponent), \
            "Wrapped component must be an ActorBaseComponent."
        self._actor = component
        # Need to record the pre-wrapped space for the wrapping functions.
        self.from_space = {
            agent.id: agent.action_space[self.key]
            for agent in self.agents.values()
            if isinstance(agent, self.supported_agent_type)
        }
        for agent in self.agents.values():
            if isinstance(agent, self.supported_agent_type):
                assert self.check_space(agent.action_space[self.key]), \
                    f"Cannot wrap {self.key} action channel for agent {agent.id}"
                agent.action_space[self.key] = self.wrap_space(agent.action_space[self.key])

    @property
    def wrapped_component(self):
        """
        Get the wrapped actor.
        """
        return self._actor

    @property
    def key(self):
        """
        The key is the same as the wrapped actor's key.
        """
        return self.wrapped_component.key

    @property
    def supported_agent_type(self):
        """
        The supported agent type is the same as the wrapped actor's supported agent type.
        """
        return self.wrapped_component.supported_agent_type

    def process_action(self, agent, action_dict, **kwargs):
        """
        Unwrap the action and pass it to the wrapped actor to process.

        Args:
            agent: The acting agent.
            action_dict: The action dictionary for this agent in this step. The
                action in this channel comes in the wrapped space.
        """
        if isinstance(agent, self.supported_agent_type):
            action = action_dict[self.key]
            unwrapped_action = self.wrap_point(self.from_space[agent.id], action)
            return self.wrapped_component.process_action(
                agent,
                {self.key: unwrapped_action},
                **kwargs
            )


# TODO Abmarl-202: Fill out the details of the abstract observer wrapper.
class ObserverWrapper(ComponentWrapper, ObserverBaseComponent):
    pass


# Docs for ObserverWrapper:
"""
Observer Wrappers
~~~~~~~~~~~~~~~~~

An Observer Wrapper uses the ``get_obs`` function, at which point
it can request an observation by passing the request to the underlying Observer
and then modify the data from the observer before sending it out. Observer Wrappers
may need to modifiy the observation spaces of corresponding agents to ensure that
the Trainer is expecting the correct format.
"""


class RavelActionWrapper(ActorWrapper):
    """
    Use numpy's ravel capabilities to convert space and points to Discrete.
    """
    def check_space(self, space):
        """
        Ensure that the space is of type that can be ravelled to discrete value.
        """
        return rdw.check_space(space)

    def wrap_space(self, space):
        """
        Convert the space into a Discrete space.
        """
        return rdw.ravel_space(space)

    def wrap_point(self, space, point):
        """
        Unravel a single discrete point to a value in the space.

        Recall that the action from the trainer arrives in the wrapped discrete
        space, so we need to unravel it so that it is in the unwrapped space before
        giving it to the actor.
        """
        return rdw.unravel(space, point)




from gym.spaces import Discrete, MultiBinary, Box, Dict, Tuple
import pytest

from abmarl.sim.agent_based_simulation import AgentBasedSimulation, Agent, PrincipleAgent

from abmarl.trainers import SinglePolicyTrainer
from abmarl.policies.policy import RandomPolicy
from abmarl.managers import AllStepManager, TurnBasedManager


class SimTest(AgentBasedSimulation):
    def __init__(self):
        self.rewards = [0, 1, 2, 3, 4, 5, 6]
        self.ith_call = -1
        self.dones = [12, 25]
        self.step_count = 0
        self.agents = {
            'agent0': Agent(
                id='agent0',
                observation_space=MultiBinary(4),
                action_space=Tuple((
                    Dict({
                        'first': Discrete(4),
                        'second': Box(low=-1, high=3, shape=(2,), dtype=int)
                    }),
                    MultiBinary(3)
                ))
            ),
            'agent1': PrincipleAgent(
                id='agent1',
            ),
            'agent2': Agent(
                id='agent2',
                observation_space=MultiBinary(4),
                action_space=Tuple((
                    Dict({
                        'first': Discrete(4),
                        'second': Box(low=-1, high=3, shape=(2,), dtype=int)
                    }),
                    MultiBinary(3)
                ))
            ),
            'agent3': PrincipleAgent(
                id='agent3',
            ),
            'agent4': PrincipleAgent(
                id='agent4'
            )
        }

    def render(self):
        pass

    def reset(self):
        self.action = {
            'agent0': None,
            'agent2': None,
        }

    def step(self, action_dict):
        self.step_count += 1
        for agent_id, action in action_dict.items():
            self.action[agent_id] = action

    def get_obs(self, agent_id, **kwargs):
        if agent_id == 'agent0':
            return [0, 0, 0, 1]
        elif agent_id == 'agent2':
            return [1, 0, 1, 0]

    def get_reward(self, agent_id, **kwargs):
        self.ith_call = (self.ith_call + 1) % 7
        return self.rewards[self.ith_call]

    def get_done(self, agent_id, **kwargs):
        if agent_id == 'agent0':
            return self.step_count >= self.dones[0]
        elif agent_id == "agent2":
            return self.step_count >= self.dones[1]

    def get_all_done(self, **kwargs):
        for agent in self.agents.values():
            if not isinstance(agent, Agent): continue
            if not self.get_done(agent.id):
                return False
        return True

    def get_info(self, agent_id, **kwargs):
        return self.action[agent_id]


class NoTrainer(SinglePolicyTrainer):
    def train(self, **kwargs):
        return self.generate_episode(horizon=20)


sim = AllStepManager(SimTest())
policy = RandomPolicy(
    action_space=sim.agents['agent0'].action_space,
    observation_space=sim.agents['agent0'].observation_space,
)


def test_trainer_sim():
    with pytest.raises(AssertionError):
        NoTrainer(sim=SimTest(), policy=policy)

    trainer = NoTrainer(sim=sim, policy=policy)

    with pytest.raises(AssertionError):
        trainer.sim = SimTest()

    trainer.sim = TurnBasedManager(SimTest())


def test_trainer_policy():
    with pytest.raises(AssertionError):
        NoTrainer(
            sim=sim,
            policy={'policy': policy}
        )
    trainer = NoTrainer(sim=sim, policy=policy)
    assert trainer.policy == policy
    assert trainer.policies == {'policy': policy}


def test_trainer_policy_mapping_function():
    trainer = NoTrainer(sim=sim, policy=policy)
    assert trainer.policy_mapping_fn('agent0') == 'policy'


def test_trainer_compute_actions():
    trainer = NoTrainer(sim=sim, policy=policy)
    obs = sim.reset()
    for agent, observation in obs.items():
        assert observation in policy.observation_space
    action = trainer.compute_actions(obs)
    assert len(action) == 2
    for agent in obs:
        assert agent in action
    for agent, act in action.items():
        assert act in policy.action_space


def test_trainer_generate_episode_policy_space_coordination():
    trainer = NoTrainer(sim=sim, policy=policy)
    observations, actions, rewards = trainer.generate_episode(horizon=20)
    for agent_id, observation in observations.items():
        assert agent_id in sim.agents
        assert type(observation) is list
        for obs in observation:
            assert obs in policy.observation_space
    for agent_id, action in actions.items():
        assert agent_id in sim.agents
        assert type(action) is list
        for act in action:
            assert act in policy.action_space
    for agent_id, reward in rewards.items():
        assert agent_id in sim.agents
        assert type(reward) is list
        for rew in reward:
            assert type(rew) in [float, int]


def test_trainer_generate_episode_check_horizon():
    trainer = NoTrainer(
        sim=sim, policy=policy
    )
    observations, actions, rewards = trainer.generate_episode(horizon=20)
    for agent_id, obs in observations.items():
        assert len(obs) <= 21
    for agent_id, action in actions.items():
        assert len(action) <= 20
    for agent_id, reward in rewards.items():
        assert len(reward) <= 20


def test_trainer_generate_episode_check_lengths():
    trainer = NoTrainer(
        sim=sim, policy=policy
    )
    observations, actions, rewards = trainer.generate_episode(horizon=20)
    for agent_id, agent in sim.agents.items():
        if not isinstance(agent, Agent): continue
        obs = observations[agent_id]
        action = actions[agent_id]
        reward = rewards[agent_id]
        assert len(obs) == len(action) + 1
        assert len(action) == len(reward)


def test_policy_action_space_mismatch():
    policy = RandomPolicy(
        action_space=Tuple((
            Dict({
                'first': Discrete(4),
            }),
            MultiBinary(3)
        )),
        observation_space=MultiBinary(4),
    )
    with pytest.raises(AssertionError):
        NoTrainer(sim=sim, policy=policy)


def test_policy_observation_space_mismatch():
    policy = RandomPolicy(
        action_space=Tuple((
            Dict({
                'first': Discrete(4),
                'second': Box(low=-1, high=3, shape=(2,), dtype=int)
            }),
            MultiBinary(3)
        )),
        observation_space=MultiBinary(5),
    )
    with pytest.raises(AssertionError):
        NoTrainer(sim=sim, policy=policy)




import pytest

from abmarl.sim import AgentBasedSimulation, PrincipleAgent, ActingAgent, ObservingAgent, Agent


def test_principle_agent_id():
    with pytest.raises(AssertionError):
        PrincipleAgent()

    with pytest.raises(AssertionError):
        PrincipleAgent(id=1)

    agent = PrincipleAgent(id='my_id')
    assert agent.id == 'my_id'
    assert agent.seed is None
    assert agent.configured

    with pytest.raises(AssertionError):
        agent.id = 4


def test_principle_agent_seed():
    with pytest.raises(AssertionError):
        PrincipleAgent(id='my_id', seed=13.5)
    agent = PrincipleAgent(id='my_id', seed=12)
    assert agent.seed == 12

    with pytest.raises(AssertionError):
        agent.seed = '12'


def test_principle_agents_equal():
    agent_1 = PrincipleAgent(id='1', seed=13)
    agent_2 = PrincipleAgent(id='1', seed=13)
    assert agent_1 == agent_2

    agent_2.id = '2'
    assert agent_1 != agent_2

    agent_2.id = '1'
    agent_2.seed = 12
    assert agent_1 != agent_2


def test_acting_agent_action_space():
    with pytest.raises(AssertionError):
        ActingAgent(id='agent', action_space=13)

    with pytest.raises(AssertionError):
        agent = ActingAgent(id='agent', action_space={'key': 'value'})

    agent = ActingAgent(id='agent')
    assert not agent.configured

    from gym.spaces import Discrete
    agent = ActingAgent(id='agent', action_space={'key': Discrete(12)})
    assert not agent.configured
    agent.finalize()
    assert agent.configured


def test_acting_agent_seed():
    from gym.spaces import Discrete
    agent = ActingAgent(id='agent', seed=17, action_space={
        1: Discrete(12),
        2: Discrete(3),
    })
    agent.finalize()
    assert agent.configured
    assert agent.action_space.sample() == {1: 5, 2: 1}


def test_observing_agent_observation_space():
    with pytest.raises(AssertionError):
        ObservingAgent(id='agent', observation_space=13)

    with pytest.raises(AssertionError):
        agent = ObservingAgent(id='agent', observation_space={'key': 'value'})

    agent = ObservingAgent(id='agent')
    assert not agent.configured

    from gym.spaces import Discrete
    agent = ObservingAgent(id='agent', observation_space={'key': Discrete(12)})
    assert not agent.configured
    agent.finalize()
    assert agent.configured


def test_agent():
    from gym.spaces import Discrete
    agent = Agent(
        id='agent', seed=7, observation_space={'obs': Discrete(2)},
        action_space={'act': Discrete(5)}
    )
    assert not agent.configured
    agent.finalize()
    assert agent.configured

    assert agent.action_space.sample() == {'act': 2}
    assert agent.observation_space.sample() == {'obs': 0}


def test_agent_based_simulation_agents():
    class ABS(AgentBasedSimulation):
        def __init__(self, agents):
            self.agents = agents

        def reset(self, **kwargs):
            pass

        def step(self, action, **kwargs):
            pass

        def render(self, **kwargs):
            pass

        def get_obs(self, agent_id, **kwargs):
            pass

        def get_reward(self, agent_id, **kwargs):
            pass

        def get_done(self, agent_id, **kwargs):
            pass

        def get_all_done(self, **kwargs):
            pass

        def get_info(self, agent_id, **kwargs):
            pass

    agents_single_object = PrincipleAgent(id='just_a_simple_agent')
    agents_list = [PrincipleAgent(id=f'{i}') for i in range(3)]
    agents_dict_key_id_no_match = {f'{i-1}': PrincipleAgent(id=f'{i}') for i in range(3)}
    agents_dict_bad_values = {f'{i}': 'PrincipleAgent(id=f"i")' for i in range(3)}
    agents_dict = {f'{i}': PrincipleAgent(id=f'{i}') for i in range(3)}

    with pytest.raises(AssertionError):
        ABS(agents=agents_single_object)

    with pytest.raises(AssertionError):
        ABS(agents=agents_list)

    with pytest.raises(AssertionError):
        ABS(agents=agents_dict_key_id_no_match)

    with pytest.raises(AssertionError):
        ABS(agents=agents_dict_bad_values)

    sim = ABS(agents=agents_dict)
    assert sim.agents == agents_dict
    sim.finalize()

    with pytest.raises(AssertionError):
        sim.agents = agents_single_object

    with pytest.raises(AssertionError):
        sim.agents = agents_list

    with pytest.raises(AssertionError):
        sim.agents = agents_dict_key_id_no_match

    with pytest.raises(AssertionError):
        ABS(agents=agents_dict_bad_values)



def create_parser(subparsers):
    analyze_parser = subparsers.add_parser('analyze', help='Analyze MARL policies')
    analyze_parser.add_argument(
        'configuration', type=str, help='Path to saved policy directory.'
    )
    analyze_parser.add_argument(
        'subscript', type=str, help='Path to subscript to run.'
    )
    analyze_parser.add_argument(
        '-c', '--checkpoint', type=int,
        help='Specify which checkpoint to load. Default is the last timestep in the directory.'
    )
    analyze_parser.add_argument('--seed', type=int, help='Seed for reproducibility.')
    return analyze_parser


def run(full_trained_directory, full_subscript, parameters):
    from abmarl import stage
    stage.run_analysis(full_trained_directory, full_subscript, parameters)



from abc import ABC, abstractmethod

from gym.spaces import Space


class Policy(ABC):
    """
    A policy maps an observation to an action.

    The underlying mapping function can take many forms. It can be Q-table, a non
    linear function, a neural network, etc.
    """
    def __init__(self, action_space=None, observation_space=None, **kwargs):
        self.action_space = action_space
        self.observation_space = observation_space

    @property
    def action_space(self):
        """
        The space of possible output actions for this policy.
        """
        return self._action_space

    @action_space.setter
    def action_space(self, value):
        assert isinstance(value, Space), "Action space must be a gym space."
        self._action_space = value

    @property
    def observation_space(self):
        """
        The space of possible input observations for this policy.
        """
        return self._observation_space

    @observation_space.setter
    def observation_space(self, value):
        assert isinstance(value, Space), "Observation space must be a gym space."
        self._observation_space = value

    @abstractmethod
    def compute_action(self, obs, **kwargs):
        """
        Generate an action from the given observation.

        Args:
            obs: The input observation.

        Returns:
            An action computed by feeding the observation forward through the underlying
            mapping.
        """
        pass

    @abstractmethod
    def update(self, obs, action, value, **kwargs):
        """
        Update the policy with the data in obs, action, and value.

        Args:
            obs: The observation involved in the update.
            action: The action involved in the update.
            value: The value to use for the update.
        """
        pass

    def reset(self):
        """
        Reset the policy at the beginning of an episode.

        Some policies behave differently at the beginning of an episode or as an episode
        progresses. The reset function allows them to reset their parameters accordingly.
        """
        pass


class RandomPolicy(Policy):
    """
    Generate random actions.
    """
    def compute_action(self, obs, **kwargs):
        """
        Check that the observation is in the space, then return a random action.

        Args:
            obs: The input observation.

        Returns:
            A random action from the action space.
        """
        assert obs in self.observation_space
        return self.action_space.sample()

    def update(self, *args, **kwargs):
        """
        RandomPolicy does not chagne, so update does nothing.
        """
        pass




import numpy as np

from abmarl.sim import PrincipleAgent, ActingAgent, ObservingAgent


class GridWorldAgent(PrincipleAgent):
    """
    The base agent in the GridWorld.
    """
    def __init__(self, initial_position=None, blocking=False, encoding=None, render_shape='o',
                 render_color='gray', **kwargs):
        super().__init__(**kwargs)
        self.encoding = encoding
        self.initial_position = initial_position
        self.blocking = blocking
        self.render_shape = render_shape
        self.render_color = render_color

    @property
    def encoding(self):
        """
        The numerical value that identifies the type of agent.

        The value does not necessarily identify the agent itself. For example,
        other agents who observe this agent will see this value.
        """
        return self._encoding

    @encoding.setter
    def encoding(self, value):
        assert type(value) is int, f"{self.id}'s encoding must be an integer."
        assert value != -2, "-2 encoding reserved for masked observation."
        assert value != -1, "-1 encoding reserved for out of bounds."
        assert value != 0, "0 encoding reserved for empty cell."
        self._encoding = value

    @property
    def initial_position(self):
        """
        The agent's initial position at reset.
        """
        return self._initial_position

    @initial_position.setter
    def initial_position(self, value):
        if value is not None:
            assert type(value) is np.ndarray, "Initial position must be a numpy array."
            assert value.shape == (2,), "Initial position must be a 2-element array."
            assert value.dtype in [int, float], "Initial position must be numerical."
        self._initial_position = value

    @property
    def position(self):
        """
        The agent's position in the grid.
        """
        return self._position

    @position.setter
    def position(self, value):
        self._position = value

    @property
    def blocking(self):
        """
        Specify if this agent blocks other agent's observations and actions.
        """
        return self._blocking

    @blocking.setter
    def blocking(self, value):
        assert type(value) is bool, "Blocking must be either True or False."
        self._blocking = value

    @property
    def render_shape(self):
        """
        The agent's shape in the rendered grid.
        """
        return self._render_shape

    @render_shape.setter
    def render_shape(self, value):
        assert value in [
            'o', 'v', '^', '<', '>', '1', '2', '3', '4', '8', 's', 'p',
            'P', '*', 'h', 'H', '+', 'x', 'X', 'D', 'd'
        ], "Invalid render shape."
        self._render_shape = value

    @property
    def render_color(self):
        """
        The agent's color in the rendered grid.
        """
        return self._render_color

    @render_color.setter
    def render_color(self, value):
        self._render_color = value

    @property
    def configured(self):
        return super().configured and self.encoding is not None and \
            self.blocking is not None and self.render_shape is not None and \
            self.render_color is not None


class GridObservingAgent(ObservingAgent, GridWorldAgent):
    """
    Observe the grid up to view range cells away.
    """
    def __init__(self, view_range=None, **kwargs):
        super().__init__(**kwargs)
        self.view_range = view_range

    @property
    def view_range(self):
        """
        The number of cells away this agent can observe in each step.
        """
        return self._view_range

    @view_range.setter
    def view_range(self, value):
        assert type(value) is int and 0 <= value, "View range must be a nonnegative integer."
        self._view_range = value

    @property
    def configured(self):
        return super().configured and self.view_range is not None


class MovingAgent(ActingAgent, GridWorldAgent):
    """
    Move up to move_range cells.
    """
    def __init__(self, move_range=None, **kwargs):
        super().__init__(**kwargs)
        self.move_range = move_range

    @property
    def move_range(self):
        """
        The maximum number of cells away that the agent can move.
        """
        return self._move_range

    @move_range.setter
    def move_range(self, value):
        assert type(value) is int and 0 <= value, "Move range must be a nonnegative integer."
        self._move_range = value

    @property
    def configured(self):
        return super().configured and self.move_range is not None


class HealthAgent(GridWorldAgent):
    """
    Agents have health points and can die.

    Health is bounded between 0 and 1.
    """
    def __init__(self, initial_health=None, **kwargs):
        super().__init__(**kwargs)
        self.initial_health = initial_health

    @property
    def health(self):
        """
        The agent's health throughout the simulation trajectory.

        The health will always be between 0 and 1.
        """
        return self._health

    @health.setter
    def health(self, value):
        assert type(value) in [int, float], "Health must be a numeric value."
        self._health = min(max(value, 0), 1)

    @property
    def initial_health(self):
        """
        The agent's initial health between 0 and 1.
        """
        return self._initial_health

    @initial_health.setter
    def initial_health(self, value):
        if value is not None:
            assert type(value) in [int, float], "Initial health must be a numeric value."
            assert 0 < value <= 1, "Initial value must be between 0 and 1."
        self._initial_health = value

    @property
    def active(self):
        """
        The agent is active if its health is greater than 0.
        """
        return self.health > 0


class AttackingAgent(ActingAgent, GridWorldAgent):
    """
    Agents that can attack other agents.
    """
    def __init__(self, attack_range=None, attack_strength=None, attack_accuracy=None, **kwargs):
        super().__init__(**kwargs)
        self.attack_range = attack_range
        self.attack_strength = attack_strength
        self.attack_accuracy = attack_accuracy

    @property
    def attack_range(self):
        """
        The maximum range of the attack.
        """
        return self._attack_range

    @attack_range.setter
    def attack_range(self, value):
        assert type(value) is int and 0 <= value, "Attack range must be a nonnegative integer."
        self._attack_range = value

    @property
    def attack_strength(self):
        """
        The strength of the attack.

        Should be between 0 and 1.
        """
        return self._attack_strength

    @attack_strength.setter
    def attack_strength(self, value):
        assert type(value) in [int, float], "Attack strength must be a numeric value."
        assert 0 <= value <= 1, "Attack strength must be between 0 and 1."
        self._attack_strength = value

    @property
    def attack_accuracy(self):
        """
        The effective accuracy of the agent's attack.

        Should be between 0 and 1. To make deterministic attacks, use 1.
        """
        return self._attack_accuracy

    @attack_accuracy.setter
    def attack_accuracy(self, value):
        assert type(value) in [int, float], "Attack accuracy must be a numeric value."
        assert 0 <= value <= 1, "Attack accuracy must be between 0 and 1."
        self._attack_accuracy = value

    @property
    def configured(self):
        return super().configured and self.attack_range is not None and \
            self.attack_strength is not None and self.attack_accuracy is not None




from gym.spaces import Box, Discrete
import numpy as np
import pytest

from abmarl.sim.gridworld.actor import MoveActor, AttackActor, ActorBaseComponent
from abmarl.sim.gridworld.state import PositionState, HealthState
from abmarl.sim.gridworld.agent import MovingAgent, AttackingAgent, HealthAgent
from abmarl.sim.gridworld.grid import Grid

from .helpers import grid


def test_move_actor():
    agents = {
        'agent0': MovingAgent(
            id='agent0', initial_position=np.array([3, 4]), encoding=1, move_range=1
        ),
        'agent1': MovingAgent(
            id='agent1', initial_position=np.array([2, 2]), encoding=2, move_range=2
        ),
        'agent2': MovingAgent(
            id='agent2', initial_position=np.array([0, 1]), encoding=1, move_range=1
        ),
        'agent3': MovingAgent(
            id='agent3', initial_position=np.array([3, 1]), encoding=3, move_range=3
        ),
    }

    position_state = PositionState(grid=grid, agents=agents)
    move_actor = MoveActor(grid=grid, agents=agents)
    assert isinstance(move_actor, ActorBaseComponent)
    assert move_actor.key == 'move'
    assert move_actor.supported_agent_type == MovingAgent
    assert agents['agent0'].action_space['move'] == Box(-1, 1, (2,), int)
    assert agents['agent1'].action_space['move'] == Box(-2, 2, (2,), int)
    assert agents['agent2'].action_space['move'] == Box(-1, 1, (2,), int)
    assert agents['agent3'].action_space['move'] == Box(-3, 3, (2,), int)

    position_state.reset()
    action = {
        'agent0': {'move': np.array([1, 1])},
        'agent1': {'move': np.array([-1, 0])},
        'agent2': {'move': np.array([0, 1])},
        'agent3': {'move': np.array([-1, 1])},
    }
    for agent_id, action in action.items():
        move_actor.process_action(agents[agent_id], action)
    np.testing.assert_array_equal(agents['agent0'].position, np.array([4, 5]))
    np.testing.assert_array_equal(agents['agent1'].position, np.array([1, 2]))
    np.testing.assert_array_equal(agents['agent2'].position, np.array([0, 2]))
    np.testing.assert_array_equal(agents['agent3'].position, np.array([2, 2]))

    action = {
        'agent0': {'move': np.array([1, 1])},
        'agent1': {'move': np.array([0, 0])},
        'agent2': {'move': np.array([-1, 1])},
        'agent3': {'move': np.array([-1, 0])},
    }
    for agent_id, action in action.items():
        move_actor.process_action(agents[agent_id], action)
    np.testing.assert_array_equal(agents['agent0'].position, np.array([4, 5]))
    np.testing.assert_array_equal(agents['agent1'].position, np.array([1, 2]))
    np.testing.assert_array_equal(agents['agent2'].position, np.array([0, 2]))
    np.testing.assert_array_equal(agents['agent3'].position, np.array([2, 2]))


def test_move_actor_with_overlap():
    overlapping = {
        1: [1],
        2: [3],
        3: [2]
    }
    grid = Grid(5, 6, overlapping=overlapping)
    agents = {
        'agent0': MovingAgent(
            id='agent0', initial_position=np.array([4, 4]), encoding=1, move_range=1
        ),
        'agent1': MovingAgent(
            id='agent1', initial_position=np.array([2, 2]), encoding=2, move_range=2
        ),
        'agent2': MovingAgent(
            id='agent2', initial_position=np.array([2, 4]), encoding=1, move_range=1
        ),
        'agent3': MovingAgent(
            id='agent3', initial_position=np.array([3, 2]), encoding=3, move_range=3
        ),
    }

    position_state = PositionState(grid=grid, agents=agents)
    move_actor = MoveActor(grid=grid, agents=agents)

    position_state.reset()
    action = {
        'agent0': {'move': np.array([-1, 0])},
        'agent1': {'move': np.array([0, 0])},
        'agent2': {'move': np.array([1, 0])},
        'agent3': {'move': np.array([-1, 0])},
    }
    for agent_id, action in action.items():
        move_actor.process_action(agents[agent_id], action)
    np.testing.assert_array_equal(agents['agent0'].position, np.array([3, 4]))
    np.testing.assert_array_equal(agents['agent1'].position, np.array([2, 2]))
    np.testing.assert_array_equal(agents['agent2'].position, np.array([3, 4]))
    np.testing.assert_array_equal(agents['agent3'].position, np.array([2, 2]))

    action = {
        'agent0': {'move': np.array([-1, 0])},
        'agent1': {'move': np.array([0, 2])},
        'agent2': {'move': np.array([0, -1])},
        'agent3': {'move': np.array([1, 1])},
    }
    for agent_id, action in action.items():
        move_actor.process_action(agents[agent_id], action)
    np.testing.assert_array_equal(agents['agent0'].position, np.array([2, 4]))
    np.testing.assert_array_equal(agents['agent1'].position, np.array([2, 2]))
    np.testing.assert_array_equal(agents['agent2'].position, np.array([3, 3]))
    np.testing.assert_array_equal(agents['agent3'].position, np.array([2, 2]))


def test_attack_actor():
    agents = {
        'agent0': HealthAgent(id='agent0', initial_position=np.array([4, 4]), encoding=1),
        'agent1': AttackingAgent(
            id='agent1',
            initial_position=np.array([2, 2]),
            encoding=1,
            attack_range=2,
            attack_strength=1,
            attack_accuracy=1
        ),
        'agent2': HealthAgent(id='agent2', initial_position=np.array([2, 3]), encoding=2),
        'agent3': HealthAgent(id='agent3', initial_position=np.array([3, 2]), encoding=1),
    }

    position_state = PositionState(grid=grid, agents=agents)
    health_state = HealthState(grid=grid, agents=agents)
    attack_actor = AttackActor(attack_mapping={1: [1]}, grid=grid, agents=agents)
    assert isinstance(attack_actor, ActorBaseComponent)
    assert attack_actor.key == 'attack'
    assert attack_actor.supported_agent_type == AttackingAgent
    assert agents['agent1'].action_space['attack'] == Discrete(2)

    position_state.reset()
    health_state.reset()
    attack_actor.process_action(agents['agent1'], {'attack': 1})
    attack_actor.process_action(agents['agent1'], {'attack': 1})
    assert not agents['agent0'].active
    assert not agents['agent3'].active
    assert agents['agent0'].health <= 0
    assert agents['agent3'].health <= 0
    assert not grid[4, 4]
    assert not grid[3, 2]

    attack_actor.process_action(agents['agent1'], {'attack': 1})
    assert agents['agent2'].active
    assert agents['agent2'].health > 0
    assert grid[2, 3]


def test_attack_actor_attack_mapping():
    agents = {
        'agent0': HealthAgent(id='agent0', initial_position=np.array([4, 4]), encoding=1),
        'agent1': AttackingAgent(
            id='agent1',
            initial_position=np.array([2, 2]),
            encoding=1,
            attack_range=2,
            attack_strength=1,
            attack_accuracy=1
        ),
        'agent2': HealthAgent(id='agent2', initial_position=np.array([2, 3]), encoding=2),
        'agent3': HealthAgent(id='agent3', initial_position=np.array([3, 2]), encoding=1),
    }

    with pytest.raises(AssertionError):
        AttackActor(agents=agents, grid=grid, attack_mapping=[1, 2, 3])

    with pytest.raises(AssertionError):
        AttackActor(agents=agents, grid=grid, attack_mapping={'1': [3], 2.0: [6]})

    with pytest.raises(AssertionError):
        AttackActor(agents=agents, grid=grid, attack_mapping={1: 3, 2: [6]})

    with pytest.raises(AssertionError):
        AttackActor(agents=agents, grid=grid, attack_mapping={1: ['2', 3], 2: [2, 3]})




from gym.spaces import MultiBinary, Discrete, Box, MultiDiscrete, Dict, Tuple
import pytest

from abmarl.sim.agent_based_simulation import AgentBasedSimulation
from abmarl.sim.agent_based_simulation import Agent, PrincipleAgent
from abmarl.sim.wrappers import SuperAgentWrapper


class SimTest(AgentBasedSimulation):
    def __init__(self):
        self.rewards = [0, 1, 2, 3, 4, 5, 6]
        self.ith_call = -1
        self.dones = [3, 12, 5, 34]
        self.step_count = 0
        self.agents = {
            'agent0': Agent(
                id='agent0',
                observation_space=MultiBinary(4),
                action_space=Tuple((
                    Dict({
                        'first': Discrete(4),
                        'second': Box(low=-1, high=3, shape=(2,), dtype=int)
                    }),
                    MultiBinary(3)
                ))
            ),
            'agent1': Agent(
                id='agent1',
                observation_space=Box(low=0, high=1, shape=(1,), dtype=int),
                action_space=MultiDiscrete([4, 6, 2])
            ),
            'agent2': Agent(
                id='agent2',
                observation_space=MultiDiscrete([2, 2]),
                action_space=Dict({'alpha': MultiBinary(3)})
            ),
            'agent3': Agent(
                id='agent3',
                observation_space=Dict({
                    'first': Discrete(4),
                    'second': Box(low=-1, high=3, shape=(2,), dtype=int)
                }),
                action_space=Tuple((Discrete(3), MultiDiscrete([10, 10]), Discrete(2)))
            ),
            'agent4': PrincipleAgent(
                id='agent4'
            )
        }

    def render(self):
        pass

    def reset(self):
        self.step_count = 0
        self.action = {agent.id: None for agent in self.agents.values() if isinstance(agent, Agent)}

    def step(self, action_dict):
        self.step_count += 1
        for agent_id, action in action_dict.items():
            self.action[agent_id] = action

    def get_obs(self, agent_id, **kwargs):
        if agent_id == 'agent0':
            return [0, 0, 0, 1]
        elif agent_id == 'agent1':
            return [0]
        elif agent_id == 'agent2':
            return [1, 0]
        elif agent_id == 'agent3':
            return {'first': 1, 'second': [3, 1]}

    def get_reward(self, agent_id, **kwargs):
        self.ith_call = (self.ith_call + 1) % 7
        return self.rewards[self.ith_call]

    def get_done(self, agent_id, **kwargs):
        return self.step_count >= self.dones[int(agent_id[-1])]

    def get_all_done(self, **kwargs):
        for agent in self.agents.values():
            if not isinstance(agent, Agent): continue
            if not self.get_done(agent.id):
                return False
        return True

    def get_info(self, agent_id, **kwargs):
        return self.action[agent_id]


sim = SuperAgentWrapper(
    SimTest(),
    super_agent_mapping={
        'super0': ['agent0', 'agent3']
    }
)
agents = sim.agents
original_agents = sim.unwrapped.agents


def test_super_agent_mapping():
    assert sim.super_agent_mapping == {
        'super0': ['agent0', 'agent3']
    }
    assert sim._covered_agents == set(('agent0', 'agent3'))
    assert sim._uncovered_agents == set(('agent1', 'agent2', 'agent4'))


def test_super_agent_mapping_breaks():
    with pytest.raises(AssertionError):
        SuperAgentWrapper(SimTest(), super_agent_mapping=['agent0'])
    with pytest.raises(AssertionError):
        SuperAgentWrapper(SimTest(), super_agent_mapping={1: ['agent0']})
    with pytest.raises(AssertionError):
        SuperAgentWrapper(SimTest(), super_agent_mapping={'agent0': ['agent1', 'agent2']})
    with pytest.raises(AssertionError):
        SuperAgentWrapper(SimTest(), super_agent_mapping={'super0': 'agent1'})
    with pytest.raises(AssertionError):
        SuperAgentWrapper(SimTest(), super_agent_mapping={'super0': [0, 1]})
    with pytest.raises(AssertionError):
        SuperAgentWrapper(SimTest(), super_agent_mapping={'super0': ['agent5']})
    with pytest.raises(AssertionError):
        SuperAgentWrapper(SimTest(), super_agent_mapping={'super0': ['agent4']})
    with pytest.raises(AssertionError):
        SuperAgentWrapper(
            SimTest(),
            super_agent_mapping={
                'super0': ['agent1', 'agent2'],
                'super1': ['agent0', 'agent1'],
            }
        )


def test_super_agent_mapping_changes_agents():
    tmp_sim = SuperAgentWrapper(
        SimTest(),
        super_agent_mapping={
            'super0': ['agent0', 'agent3']
        }
    )
    assert tmp_sim.agents.keys() == set(('super0', 'agent1', 'agent2', 'agent4'))
    assert 'agent0' in tmp_sim.agents['super0'].action_space.spaces
    assert 'agent3' in tmp_sim.agents['super0'].action_space.spaces
    tmp_sim.super_agent_mapping = {
        'super0': ['agent1', 'agent0'],
        'super1': ['agent2', 'agent3']
    }
    assert tmp_sim.agents.keys() == set(('super0', 'super1', 'agent4'))
    assert 'agent0' in tmp_sim.agents['super0'].action_space.spaces
    assert 'agent1' in tmp_sim.agents['super0'].action_space.spaces
    assert 'agent2' in tmp_sim.agents['super1'].action_space.spaces
    assert 'agent3' in tmp_sim.agents['super1'].action_space.spaces


def test_agent_spaces():
    assert len(agents) == 4
    assert agents['super0'].action_space == Dict({
        'agent0': original_agents['agent0'].action_space,
        'agent3': original_agents['agent3'].action_space,
    })
    assert agents['super0'].observation_space == Dict({
        'agent0': original_agents['agent0'].observation_space,
        'agent3': original_agents['agent3'].observation_space,
        'mask': Dict({'agent0': Discrete(2), 'agent3': Discrete(2)})
    })
    assert agents['agent1'] == original_agents['agent1']
    assert agents['agent2'] == original_agents['agent2']
    assert agents['agent4'] == original_agents['agent4']


def test_sim_step():
    sim.reset()
    assert sim.unwrapped.action == {
        'agent0': None,
        'agent1': None,
        'agent2': None,
        'agent3': None
    }

    actions = {
        'super0': {
            'agent0': ({'first': 2, 'second': [-1, 2]}, [0, 1, 0]),
            'agent3': (0, [7, 3], 1)
        },
        'agent1': [2, 3, 0],
        'agent2': {'alpha': [1, 1, 1]}
    }
    sim.step(actions)
    assert sim.unwrapped.action == {
        'agent0': ({'first': 2, 'second': [-1, 2]}, [0, 1, 0]),
        'agent3': (0, [7, 3], 1),
        'agent1': [2, 3, 0],
        'agent2': {'alpha': [1, 1, 1]}
    }


def test_sim_step_breaks():
    actions = {
        'agent0': ({'first': 2, 'second': [-1, 2]}, [0, 1, 0]),
        'agent3': (0, [7, 3], 1),
        'agent1': [2, 3, 0],
        'agent2': {'alpha': [1, 1, 1]}
    }
    with pytest.raises(AssertionError):
        sim.step(actions)


def test_sim_step_covered_agent_done():
    sim.reset()
    sim.unwrapped.step_count = 4
    actions = {
        'super0': {
            'agent0': ({'first': 2, 'second': [-1, 2]}, [0, 1, 0]),
            'agent3': (0, [7, 3], 1)
        },
        'agent1': [2, 3, 0],
        'agent2': {'alpha': [1, 1, 1]}
    }
    sim.step(actions)
    assert sim.unwrapped.action == {
        'agent0': None,
        'agent3': (0, [7, 3], 1),
        'agent1': [2, 3, 0],
        'agent2': {'alpha': [1, 1, 1]}
    }


def test_sim_obs():
    sim.unwrapped.step_count = 4
    obs = sim.get_obs('super0')
    assert obs in agents['super0'].observation_space
    assert obs == {
        'agent0': [0, 0, 0, 1],
        'agent3': {'first': 1, 'second': [3, 1]},
        'mask': {'agent0': False, 'agent3': True}
    }

    obs = sim.get_obs('agent1')
    assert obs in agents['agent1'].observation_space
    assert obs == [0]

    obs = sim.get_obs('agent2')
    assert obs in agents['agent2'].observation_space
    assert obs == [1, 0]


def test_sim_obs_breaks():
    with pytest.raises(AssertionError):
        sim.get_obs('agent0')
    with pytest.raises(AssertionError):
        sim.get_obs('agent3')


def test_sim_rewards():
    assert sim.get_reward('super0') == 1
    assert sim.get_reward('agent1') == 2
    assert sim.get_reward('agent2') == 3
    assert sim.get_reward('super0') == 9


def test_sim_rewards_breaks():
    with pytest.raises(AssertionError):
        sim.get_reward('agent0')
    with pytest.raises(AssertionError):
        sim.get_reward('agent3')


def test_sim_done():
    sim.unwrapped.step_count = 10
    assert not sim.get_done('super0')
    assert not sim.get_done('agent1')
    assert sim.get_done('agent2')

    sim.unwrapped.step_count = 40
    assert sim.get_done('super0')
    assert sim.get_done('agent1')


def test_sim_done_breaks():
    with pytest.raises(AssertionError):
        sim.get_done('agent0')
    with pytest.raises(AssertionError):
        sim.get_done('agent3')


def test_sim_all_done():
    sim.unwrapped.step_count = 15
    assert not sim.get_all_done()
    sim.unwrapped.step_count = 40
    assert sim.get_all_done()


def test_double_wrap():
    sim2 = SuperAgentWrapper(
        sim,
        super_agent_mapping={
            'double0': ['super0', 'agent1']
        }
    )
    assert sim2._covered_agents == set(('super0', 'agent1'))
    assert sim2._uncovered_agents == set(('agent2', 'agent4'))
    assert sim2.agents.keys() == set(('double0', 'agent2', 'agent4'))
    assert 'super0' in sim2.agents['double0'].action_space.spaces
    assert 'agent1' in sim2.agents['double0'].action_space.spaces

    assert len(sim2.agents) == 3
    assert sim2.agents['double0'].action_space == Dict({
        'super0': sim2.sim.agents['super0'].action_space,
        'agent1': sim2.sim.agents['agent1'].action_space
    })
    assert sim2.agents['double0'].observation_space == Dict({
        'super0': sim2.sim.agents['super0'].observation_space,
        'agent1': sim2.sim.agents['agent1'].observation_space,
        'mask': Dict({'super0': Discrete(2), 'agent1': Discrete(2)})
    })
    assert sim2.agents['agent2'] == original_agents['agent2']
    assert sim2.agents['agent4'] == original_agents['agent4']

    sim2.reset()
    assert sim2.unwrapped.action == {
        'agent0': None,
        'agent1': None,
        'agent2': None,
        'agent3': None
    }
    actions = {
        'double0': {
            'super0': {
                'agent0': ({'first': 2, 'second': [-1, 2]}, [0, 1, 0]),
                'agent3': (0, [7, 3], 1)
            },
            'agent1': [2, 3, 0],
        },
        'agent2': {'alpha': [1, 1, 1]}
    }
    sim2.step(actions)
    assert sim2.unwrapped.action == {
        'agent0': ({'first': 2, 'second': [-1, 2]}, [0, 1, 0]),
        'agent3': (0, [7, 3], 1),
        'agent1': [2, 3, 0],
        'agent2': {'alpha': [1, 1, 1]}
    }

    obs = sim2.get_obs('double0')
    assert obs in sim2.agents['double0'].observation_space
    assert obs == {
        'super0': {
            'agent0': [0, 0, 0, 1],
            'agent3': {'first': 1, 'second': [3, 1]},
            'mask': {'agent0': True, 'agent3': True}
        },
        'agent1': [0],
        'mask': {'agent1': True, 'super0': True}
    }
    obs = sim2.get_obs('agent2')
    assert obs in sim2.agents['agent2'].observation_space
    assert obs == [1, 0]

    sim2.unwrapped.ith_call = -1
    assert sim2.get_reward('double0') == 3
    assert sim2.get_reward('agent2') == 3
    assert sim2.get_reward('double0') == 15

    sim2.unwrapped.step_count = 4
    assert not sim2.get_done('double0')
    assert not sim2.get_done('agent2')

    sim2.unwrapped.step_count = 32
    assert not sim2.get_done('double0')
    assert sim2.get_done('agent2')

    sim2.unwrapped.step_count = 40
    assert sim2.get_done('double0')
    assert sim2.get_done('agent2')



from abmarl.sim.corridor import MultiCorridor
from abmarl.managers import TurnBasedManager
from abmarl.external import MultiAgentWrapper

sim = MultiAgentWrapper(TurnBasedManager(MultiCorridor()))

sim_name = "MultiCorridor"
from ray.tune.registry import register_env
register_env(sim_name, lambda sim_config: sim)

ref_agent = sim.sim.agents['agent0']
policies = {
    'corridor': (None, ref_agent.observation_space, ref_agent.action_space, {})
}


def policy_mapping_fn(agent_id):
    return 'corridor'


# Experiment parameters
params = {
    'experiment': {
        'title': f'{sim_name}',
        'sim_creator': lambda config=None: sim,
    },
    'ray_tune': {
        'run_or_experiment': 'PG',
        'checkpoint_freq': 50,
        'checkpoint_at_end': True,
        'stop': {
            'episodes_total': 2000,
        },
        'verbose': 2,
        'config': {
            # --- Simulation ---
            'disable_env_checking': True,
            'env': sim_name,
            'horizon': 200,
            'env_config': {},
            # --- Multiagent ---
            'multiagent': {
                'policies': policies,
                'policy_mapping_fn': policy_mapping_fn,
            },
            # "lr": 0.0001,
            # --- Parallelism ---
            # Number of workers per experiment: int
            "num_workers": 7,
            # Number of simulations that each worker starts: int
            "num_envs_per_worker": 1, # This must be 1 because we are not "threadsafe"
        },
    }
}



def create_parser(subparsers):
    visualize_parser = subparsers.add_parser('visualize', help='Visualize MARL policies')
    visualize_parser.add_argument(
        'configuration', type=str, help='Path to saved policy directory.'
    )
    visualize_parser.add_argument(
        '-c', '--checkpoint', type=int,
        help='Specify which checkpoint to load. Default is the last timestep in the directory.'
    )
    visualize_parser.add_argument(
        '-n', '--episodes', type=int, default=1, help='The number of episodes to run. Default 1.'
    )
    visualize_parser.add_argument(
        '-s', '--steps-per-episode', type=int, default=200,
        help='The maximum number of steps to take per epsiode. Default 200.'
    )
    visualize_parser.add_argument(
        '--record', action='store_true',
        help='Record a video of the agent(s) interacting in the simulation.'
    )
    visualize_parser.add_argument(
        '--frame-delay', type=int,
        help='The number of milliseconds to delay between each frame in the animation.',
        default=200
    )
    visualize_parser.add_argument(
        '--no-explore', action='store_false', help='Turn off exploration in the action policy.'
    )
    visualize_parser.add_argument('--seed', type=int, help='Seed for reproducibility.')
    return visualize_parser


def run(full_trained_directory, parameters):
    from abmarl import stage
    stage.run_visualize(full_trained_directory, parameters)




from abc import ABC, abstractmethod

import numpy as np
from gym.spaces import Box, Discrete

from abmarl.sim.gridworld.base import GridWorldBaseComponent
from abmarl.sim.gridworld.agent import MovingAgent, AttackingAgent
import abmarl.sim.gridworld.utils as gu


class ActorBaseComponent(GridWorldBaseComponent, ABC):
    """
    Abstract Actor Component class from which all Actor Components will inherit.
    """
    @abstractmethod
    def process_action(self, agent, action_dict, **kwargs):
        """
        Process the agent's action.

        Args:
            agent: The acting agent.
            action_dict: The action dictionary for this agent in this step. The
                dictionary may have different entries, each of which will be processed
                by different Actors.
        """
        pass

    @property
    @abstractmethod
    def key(self):
        """
        The key in the action dictionary.

        The action space of all acting agents in the gridworld framework is a dict.
        We can build up complex action spaces with multiple components by
        assigning each component an entry in the action dictionary. Actions
        will be a dictionary even if your simulation only has one Actor.
        """
        pass

    @property
    @abstractmethod
    def supported_agent_type(self):
        """
        The type of Agent that this Actor works with.

        If an agent is this type, the Actor will add its entry to the
        agent's action space and will process actions for this agent.
        """
        pass


class MoveActor(ActorBaseComponent):
    """
    Agents can move to unoccupied nearby squares.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        for agent in self.agents.values():
            if isinstance(agent, self.supported_agent_type):
                agent.action_space[self.key] = Box(
                    -agent.move_range, agent.move_range, (2,), int
                )

    @property
    def key(self):
        """
        This Actor's key is "move".
        """
        return "move"

    @property
    def supported_agent_type(self):
        """
        This Actor works with MovingAgents.
        """
        return MovingAgent

    def process_action(self, agent, action_dict, **kwargs):
        """
        The agent can move to nearby squares.

        The agent's new position must be within the grid and the cell-occupation rules
        must be met.

        Args:
            agent: Move the agent if it is a MovingAgent.
            action_dict: The action dictionary for this agent in this step. If
                the agent is a MovingAgent, then the action dictionary will contain
                the "move" entry.

        Returns:
            True if the move is successful, False otherwise.
        """
        if isinstance(agent, self.supported_agent_type):
            action = action_dict[self.key]
            new_position = agent.position + action
            if 0 <= new_position[0] < self.rows and \
                    0 <= new_position[1] < self.cols:
                from_ndx = tuple(agent.position)
                to_ndx = tuple(new_position)
                if to_ndx == from_ndx:
                    return True
                elif self.grid.query(agent, to_ndx):
                    self.grid.remove(agent, from_ndx)
                    self.grid.place(agent, to_ndx)
                    return True
                else:
                    return False
            else:
                return False


class AttackActor(ActorBaseComponent):
    """
    Agents can attack other agents.
    """
    def __init__(self, attack_mapping=None, **kwargs):
        super().__init__(**kwargs)
        self.attack_mapping = attack_mapping
        for agent in self.agents.values():
            if isinstance(agent, self.supported_agent_type):
                agent.action_space[self.key] = Discrete(2)

    @property
    def attack_mapping(self):
        """
        Dict that dictates which agents the attacking agent can attack.

        The dictionary maps the attacking agents' encodings to a list of encodings
        that they can attack.
        """
        return self._attack_mapping

    @attack_mapping.setter
    def attack_mapping(self, value):
        assert type(value) is dict, "Attack mapping must be dictionary."
        for k, v in value.items():
            assert type(k) is int, "All keys in attack mapping must be integer."
            assert type(v) is list, "All values in attack mapping must be list."
            for i in v:
                assert type(i) is int, \
                    "All elements in the attack mapping values must be integers."
        self._attack_mapping = value

    @property
    def key(self):
        """
        This Actor's key is "attack".
        """
        return 'attack'

    @property
    def supported_agent_type(self):
        """
        This Actor works with AttackingAgents.
        """
        return AttackingAgent

    def process_action(self, attacking_agent, action_dict, **kwargs):
        """
        If the agent has chosen to attack, then we process their attack.

        The processing goes through a series of checks. The attack is possible
        if there is an attacked agent such that:

        1. The attacked agent is active.
        2. The attacked agent is within range.
        3. The attacked agent is valid according to the attack_mapping.

        If the attack is possible, then we determine the success of the attack
        based on the attacking agent's accuracy. If the attack is successful, then
        the attacked agent's health is depleted by the attacking agent's strength,
        possibly resulting in its death.
        """
        def determine_attack(agent):
            # Generate local grid and an attack mask.
            local_grid, mask = gu.create_grid_and_mask(
                agent, self.grid, agent.attack_range, self.agents
            )

            # Randomly scan the local grid for attackable agents.
            attackable_agents = []
            for r in range(2 * agent.attack_range + 1):
                for c in range(2 * agent.attack_range + 1):
                    if mask[r, c]: # We can see this cell
                        candidate_agents = local_grid[r, c]
                        if candidate_agents is not None:
                            for other in candidate_agents.values():
                                if other.id == agent.id: # Cannot attack yourself
                                    continue
                                elif not other.active: # Cannot attack inactive agents
                                    continue
                                elif other.encoding not in self.attack_mapping[agent.encoding]:
                                    # Cannot attack this type of agent
                                    continue
                                elif np.random.uniform() > agent.attack_accuracy:
                                    # Failed attack
                                    continue
                                else:
                                    attackable_agents.append(other)
            return np.random.choice(attackable_agents) if attackable_agents else None

        if isinstance(attacking_agent, self.supported_agent_type):
            action = action_dict[self.key]
            if action: # Agent has chosen to attack
                attacked_agent = determine_attack(attacking_agent)
                if attacked_agent is not None:
                    attacked_agent.health = attacked_agent.health - attacking_agent.attack_strength
                    if not attacked_agent.active:
                        self.grid.remove(attacked_agent, attacked_agent.position)
                return attacked_agent




import numpy as np

from abmarl.sim.gridworld.examples.team_battle_example import BattleAgent, TeamBattleSim
from abmarl.managers import AllStepManager
from abmarl.external import MultiAgentWrapper


colors = ['red', 'blue', 'green', 'gray']
positions = [np.array([1, 1]), np.array([1, 6]), np.array([6, 1]), np.array([6, 6])]
agents = {
    f'agent{i}': BattleAgent(
        id=f'agent{i}',
        encoding=i % 4 + 1,
        render_color=colors[i % 4],
        initial_position=positions[i % 4]
    ) for i in range(24)
}
overlap_map = {
    1: [1],
    2: [2],
    3: [3],
    4: [4]
}
attack_map = {
    1: [2, 3, 4],
    2: [1, 3, 4],
    3: [1, 2, 4],
    4: [1, 2, 3]
}
sim = MultiAgentWrapper(
    AllStepManager(
        TeamBattleSim.build_sim(
            8, 8,
            agents=agents,
            overlapping=overlap_map,
            attack_mapping=attack_map
        )
    )
)


sim_name = "TeamBattle"
from ray.tune.registry import register_env
register_env(sim_name, lambda sim_config: sim)


policies = {
    'red': (None, agents['agent0'].observation_space, agents['agent0'].action_space, {}),
    'blue': (None, agents['agent1'].observation_space, agents['agent1'].action_space, {}),
    'green': (None, agents['agent2'].observation_space, agents['agent2'].action_space, {}),
    'gray': (None, agents['agent3'].observation_space, agents['agent3'].action_space, {}),
}


def policy_mapping_fn(agent_id):
    if agents[agent_id].encoding == 1:
        return 'red'
    if agents[agent_id].encoding == 2:
        return 'blue'
    if agents[agent_id].encoding == 3:
        return 'green'
    if agents[agent_id].encoding == 4:
        return 'gray'


# Experiment parameters
params = {
    'experiment': {
        'title': f'{sim_name}',
        'sim_creator': lambda config=None: sim,
    },
    'ray_tune': {
        'run_or_experiment': 'A2C',
        'checkpoint_freq': 50,
        'checkpoint_at_end': True,
        'stop': {
            'episodes_total': 2000,
        },
        'verbose': 2,
        'config': {
            # --- Simulation ---
            'disable_env_checking': False,
            'env': sim_name,
            'horizon': 200,
            'env_config': {},
            # --- Multiagent ---
            'multiagent': {
                'policies': policies,
                'policy_mapping_fn': policy_mapping_fn,
            },
            # "lr": 0.0001,
            # --- Parallelism ---
            # Number of workers per experiment: int
            "num_workers": 7,
            # Number of simulations that each worker starts: int
            "num_envs_per_worker": 1, # This must be 1 because we are not "threadsafe"
        },
    }
}




from abc import ABC, abstractmethod

from abmarl.policies.policy import Policy
from abmarl.managers import SimulationManager
from abmarl.sim.agent_based_simulation import Agent


class MultiPolicyTrainer(ABC):
    """
    Train policies with data generated by agents interacting in a simulation.
    """
    def __init__(self, sim=None, policies=None, policy_mapping_fn=None, **kwargs):
        self.sim = sim
        self.policies = policies
        self.policy_mapping_fn = policy_mapping_fn
        self._check_agent_policy_alignment()

    @property
    def sim(self):
        """
        The SimulationManager.
        """
        return self._sim

    @sim.setter
    def sim(self, value):
        assert isinstance(value, SimulationManager), "sim must be a Simulation Manager."
        self._sim = value

    @property
    def policies(self):
        """
        A dictionary that maps the policy id's to a policy object.
        """
        return self._policies

    @policies.setter
    def policies(self, value):
        assert type(value) is dict, "Policies must be a dict"
        for k, v in value.items():
            assert type(k) is str, \
                "The keys in the policies dictionary must be the policy names as strings."
            assert isinstance(v, Policy), \
                "The values in the policies dictionary must be Policy objects."
        self._policies = value

    @property
    def policy_mapping_fn(self):
        """
        A function that takes an agent's id as input and outputs its corresponding policy id.
        """
        return self._policy_mapping_fn

    @policy_mapping_fn.setter
    def policy_mapping_fn(self, value):
        assert callable(value), "Policy Mapping Function must be a function."
        self._policy_mapping_fn = value

    def compute_actions(self, obs):
        """
        Compute actions for agents in the observation.

        Forwards the observations to the respective policy for each agent that
        reports an observation.

        Args:
            obs: an observation dictionary, where the keys are the agents reporting
                from the sim and the values are the observations.

        Returns:
            An action dictionary where the keys are the agents from the observation
                and the values are the actions generated from each agent's policy.
        """
        return {
            agent_id: self.policies[self.policy_mapping_fn(agent_id)].compute_action(obs[agent_id])
            for agent_id in obs
        }

    def generate_episode(self, horizon=200, **kwargs):
        """
        Generate an episode of data.

        The fundamental data object is a SAR--a (state, action, reward) tuple.
        We restart the sim, generating initial observations (states) for agents
        reporting from the sim. Then we use the compute_action function to generate
        actions for agents who report an observation. Those actions are given to the sim,
        which steps forward and generates rewards and new observations for reporting
        agents. This loop continues until the simulation is done or we hit the
        horizon.

        Args:
            horizon: The maximum number of steps per epsidoe. The episode may
                finish early, but it will not progress further than this number
                of steps.

        Returns:
            Three dictionaries: one for observations, another for actions, and
            another for rewards, thus making up the SAR sequences. The data is
            organized by agent_id, so you would call
                {observations, actions, rewards}[agent_id][i]
            in order to extract the ith SAR for an agent.
            NOTE: In multiagent simulations, the number of SARs may differ for
            each agent.
        """
        # Reset the simulation and policies
        obs = self.sim.reset()
        done = {agent: False for agent in obs}
        for policy in self.policies.values():
            policy.reset()

        # Data collection
        observations, actions, rewards = {}, {}, {}
        for agent_id, agent_obs in obs.items():
            observations[agent_id] = [agent_obs]

        # Generate episode of data
        for j in range(horizon):
            action = self.compute_actions(obs)
            obs, reward, done, _ = self.sim.step(action)

            # Store the data
            for agent_id, agent_obs in obs.items():
                try:
                    observations[agent_id].append(agent_obs)
                except KeyError:
                    observations[agent_id] = [agent_obs]
            for agent_id, agent_reward in reward.items():
                try:
                    rewards[agent_id].append(agent_reward)
                except KeyError:
                    rewards[agent_id] = [agent_reward]
            for agent_id, agent_action in action.items():
                try:
                    actions[agent_id].append(agent_action)
                except KeyError:
                    actions[agent_id] = [agent_action]

            # Exit if we're done
            if done['__all__']:
                break

            # We should not request actions for any more done agents.
            for agent_id, agent_done in done.items():
                if agent_done:
                    # This is okay to delete because it has already been stored in observations
                    del obs[agent_id]

        return observations, actions, rewards

    @abstractmethod
    def train(self, iterations=10_000, **kwargs):
        """
        Train the policy objects using generated data.

        This function is abstract and should be implemented by the algorithm.
        The implementation should look something like this:
        for iter in range(iterations):
            observations, actions, rewards = self.generate_episode()
            # Implementation: update the policy with the generated data.

        Args:
            iterations: The number of training iterations.
            **kwargs: Any additional parameter your algorithm may need.
        """
        pass

    def _check_agent_policy_alignment(self):
        """
        Check agent-policy alignment.

        Check that every agent that acts and observes is assigned a policy and
        that the action and observation spaces between each agent and its assigned
        policy align.
        """
        # Quick assertion that all the spaces lines up
        for agent in self.sim.agents.values():
            if not isinstance(agent, Agent): continue
            policy_id = self.policy_mapping_fn(agent.id)
            policy = self.policies[policy_id]
            assert agent.action_space == policy.action_space, \
                f"agent{agent.id} has been assigned to policy {policy_id} but " + \
                "the action spaces are different."
            assert agent.observation_space == policy.observation_space, \
                f"agent{agent.id} has been assigned to policy {policy_id} but " + \
                "the observation spaces are different."


class SinglePolicyTrainer(MultiPolicyTrainer):
    """
    Train a single policy with data generated by agents interacting in a simulation.
    """
    def __init__(self, sim=None, policy=None, **kwargs):
        self.sim = sim
        self.policy = policy
        self._check_agent_policy_alignment()

    @property
    def policy(self):
        """
        The policy to train.
        """
        return self._policy

    @policy.setter
    def policy(self, value):
        assert isinstance(value, Policy), "policy must be a Policy object."
        self._policy = value

    @property
    def policies(self):
        return {'policy': self.policy}

    @property
    def policy_mapping_fn(self):
        """
        Return function always returns "policy", which is the name we give the policy.
        """
        return lambda agent_id: 'policy'



from abmarl.sim import ObservingAgent, ActingAgent

from .simulation_manager import SimulationManager


class AllStepManager(SimulationManager):
    """
    The AllStepManager gets the observations of all agents at reset. At step, it gets
    the observations of all the agents that are not done. Once all the agents
    are done, the manager returns all done.
    """
    def reset(self, **kwargs):
        """
        Reset the simulation and return the observation of all the agents.
        """
        self.done_agents = set(
            agent.id for agent in self.agents.values()
            if not (isinstance(agent, ActingAgent) and isinstance(agent, ObservingAgent))
        )
        self.sim.reset(**kwargs)
        return {
            agent.id: self.sim.get_obs(agent.id)
            for agent in self.agents.values() if agent.id not in self.done_agents
        }

    def step(self, action_dict, **kwargs):
        """
        Assert that the incoming action does not come from an agent who is recorded
        as done. Step the simulation forward and return the observation, reward,
        done, and info of all the non-done agents, including the agents that were
        done in this step. If all agents are done in this turn, then the manager
        returns all done.
        """
        for agent_id in action_dict:
            assert agent_id not in self.done_agents, \
                "Received an action for an agent that is already done."
        self.sim.step(action_dict, **kwargs)

        obs = {
            agent.id: self.sim.get_obs(agent.id) for agent in self.agents.values()
            if agent.id not in self.done_agents
        }
        rewards = {
            agent.id: self.sim.get_reward(agent.id) for agent in self.agents.values()
            if agent.id not in self.done_agents
        }
        dones = {
            agent.id: self.sim.get_done(agent.id) for agent in self.agents.values()
            if agent.id not in self.done_agents
        }
        infos = {
            agent.id: self.sim.get_info(agent.id) for agent in self.agents.values()
            if agent.id not in self.done_agents
        }

        for agent, done in dones.items():
            if done:
                self.done_agents.add(agent)

        # if all agents are done or the simulation is done, then return done
        if self.sim.get_all_done() or not (self.agents.keys() - self.done_agents):
            dones['__all__'] = True
        else:
            dones['__all__'] = False

        return obs, rewards, dones, infos



from .wrapper import Wrapper


class SARWrapper(Wrapper):
    """
    Wraps the actions and observations for all the agents at reset and step.
    To create your own wrapper, inherit from this class and override the wrap
    and unwrap functions.

    Note: wrapping the action "goes the other way" than the reward and observation, like this:
        obs:    sim agent -> wrapper -> trainer
        reward: sim agent -> wrapper -> trainer
        action: sim agent <- wrapper <- trainer

    If you wrap an action, be aware that the wrapper must return what the simulation
    agents expect; whereas if you wrap an observation or reward, the wrapper must return
    what the trainer expects. The expectations are defined by the observation and
    action spaces of the wrapped simulation agents at initialization.
    """
    def step(self, action_dict, **kwargs):
        """
        Wrap each of the agent's actions from the policies before passing them
        to sim.step.
        """
        self.sim.step(
            {
                agent_id: self.wrap_action(self.sim.agents[agent_id], action)
                for agent_id, action in action_dict.items()
            },
            **kwargs
        )

    def get_obs(self, agent_id, **kwargs):
        return self.wrap_observation(self.sim.agents[agent_id], self.sim.get_obs(agent_id))

    def get_reward(self, agent_id, **kwargs):
        return self.wrap_reward(self.sim.get_reward(agent_id))

    # Default wrapping and unwrapping behavior. Override these in your custom wrapper.
    # Developer note: we have to have separate wrappers for each because we don't
    # want to force the observation and action space to map to the same wrapped space.
    def wrap_observation(self, from_agent, observation):
        return observation

    def unwrap_observation(self, from_agent, observation):
        return observation

    def wrap_action(self, from_agent, action):
        return action

    def unwrap_action(self, from_agent, action):
        return action

    def wrap_reward(self, reward):
        return reward

    def unwrap_reward(self, reward):
        return reward



# 5.1 Monte Carlo prediction p92.
from gym.spaces import Discrete
import numpy as np

from abmarl.managers import SimulationManager
from abmarl.external import GymWrapper
from abmarl.policies.q_table_policy import EpsilonSoftPolicy


def generate_episode(sim, policy, horizon=200):
    """
    Generate an episode from a policy acting on an simulation.

    Returns: sequence of state, action, reward.
    """
    obs = sim.reset()
    policy.reset() # Reset the policy too so that it knows its the beginning of the episode.
    states, actions, rewards = [], [], []
    states.append(obs)
    for _ in range(horizon):
        action = policy.compute_action(obs)
        obs, reward, done, _ = sim.step(action)
        states.append(obs)
        actions.append(action)
        rewards.append(reward)
        if done:
            break

    states.pop() # Pop off the terminating state
    return states, actions, rewards


def off_policy(sim, iteration=10_000, gamma=0.9, horizon=200):
    """
    Off-policy Monte Carlo control estimates an optimal policy in a simulation. Trains a greedy
    policy be generating trajectories from an epsilon-soft behavior policy.

    Args:
        sim: The simulation, obviously.
        iteration: The number of times to iterate the learning algorithm.
        gamme: The discount factor

    Returns:
        sim: The simulation. Algorithms may wrap simulations before training in them, so this
            simulation may be wrapped.
        q_table: The Q values
        policy: The policy that is learned.
    """
    assert isinstance(sim, SimulationManager)
    sim = GymWrapper(sim)
    assert isinstance(sim.observation_space, Discrete)
    assert isinstance(sim.action_space, Discrete)
    q_table = np.random.normal(0, 1, size=(sim.observation_space.n, sim.action_space.n))
    c_table = 0 * q_table
    policy = EpsilonSoftPolicy(
        observation_space=sim.observation_space,
        action_space=sim.action_space,
        q_table=q_table
    )
    for i in range(iteration):
        behavior_policy = EpsilonSoftPolicy(
            observation_space=policy.observation_space,
            action_space=policy.action_space,
            q_table=policy.q_table
        )
        states, actions, rewards, = generate_episode(sim, behavior_policy, horizon)
        G = 0
        W = 1
        for i in reversed(range(len(states))):
            state, action, reward = states[i], actions[i], rewards[i]
            G = gamma * G + reward
            c_table[state, action] += W
            q_table[state, action] = q_table[state, action] + W/(c_table[state, action]) * \
                (G - q_table[state, action])
            if action != policy.compute_action(state): # Nonoptimal action
                break
            W /= behavior_policy.probability(state, action)

    return sim, q_table, policy



from abmarl.tools import utils as adu


def run(full_config_path):
    """Train MARL policies using the config_file."""

    # Load the experiment as a module
    experiment_mod = adu.custom_import_module(full_config_path)
    title = experiment_mod.params['experiment']['title']

    # Copy the configuration module to the output directory
    import os
    import shutil
    import time
    base = experiment_mod.params['ray_tune'].get('local_dir', os.path.expanduser("~"))
    output_dir = os.path.join(
        base, 'abmarl_results/{}_{}'.format(
            title, time.strftime('%Y-%m-%d_%H-%M')
        )
    )
    experiment_mod.params['ray_tune']['local_dir'] = output_dir
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    shutil.copy(full_config_path, output_dir)

    # Train with ray
    import ray
    from ray import tune
    ray.init()
    tune.run(**experiment_mod.params['ray_tune'])
    ray.shutdown()



def create_parser(subparsers):
    """Parse the arguments for the train command.

    Returns
    -------
        parser : ArgumentParser
    """
    train_parser = subparsers.add_parser('train', help='Train MARL policies ')
    train_parser.add_argument(
        'configuration', type=str, help='Path to python config file. Include the .py extension.'
    )
    return train_parser


def run(full_config_path):
    from abmarl import train
    train.run(full_config_path)




import numpy as np
import pytest

from abmarl.sim.gridworld.agent import GridWorldAgent, GridObservingAgent, MovingAgent, \
    HealthAgent, AttackingAgent
from abmarl.sim import PrincipleAgent, ActingAgent, ObservingAgent


def test_grid_world_agent():
    agent = GridWorldAgent(
        id='agent',
        initial_position=np.array([2, 2]),
        blocking=True,
        encoding=4
    )
    assert isinstance(agent, PrincipleAgent)
    assert agent.id == 'agent'
    np.testing.assert_array_equal(agent.initial_position, np.array([2, 2]))
    assert agent.blocking
    assert agent.encoding == 4
    assert agent.render_shape == 'o'
    assert agent.render_color == 'gray'
    assert agent.configured

    # Encoding
    with pytest.raises(AssertionError):
        agent = GridWorldAgent(
            id='agent',
            encoding='2'
        )
    with pytest.raises(AssertionError):
        agent = GridWorldAgent(
            id='agent',
            encoding=-2
        )
    with pytest.raises(AssertionError):
        agent = GridWorldAgent(
            id='agent',
            encoding=-1
        )
    with pytest.raises(AssertionError):
        agent = GridWorldAgent(
            id='agent',
            encoding=0
        )

    # Initial position
    with pytest.raises(AssertionError):
        agent = GridWorldAgent(
            id='agent',
            initial_position=np.array(['0', '1'])
        )
    with pytest.raises(AssertionError):
        agent = GridWorldAgent(
            id='agent',
            initial_position=np.array([0, 1, 2])
        )
    with pytest.raises(AssertionError):
        agent = GridWorldAgent(
            id='agent',
            initial_position=[0, 1, 2]
        )

    # Blocking
    with pytest.raises(AssertionError):
        agent = GridWorldAgent(
            id='agent',
            blocking=1
        )

    # Render shape
    with pytest.raises(AssertionError):
        agent = GridWorldAgent(
            id='agent',
            render_shape='circle'
        )


def test_grid_observing_agent():
    agent = GridObservingAgent(
        id='agent',
        encoding=1,
        view_range=3
    )
    assert isinstance(agent, ObservingAgent)
    assert isinstance(agent, GridWorldAgent)
    assert agent.view_range == 3

    # View range
    with pytest.raises(AssertionError):
        agent = GridObservingAgent(
            id='agent',
            encoding=1,
            view_range=-1
        )


def test_moving_agent():
    agent = MovingAgent(
        id='agent',
        encoding=1,
        move_range=2
    )
    assert isinstance(agent, ActingAgent)
    assert isinstance(agent, GridWorldAgent)
    assert agent.move_range == 2

    with pytest.raises(AssertionError):
        agent = MovingAgent(
            id='agent',
            encoding=1,
            move_range='1'
        )


def test_health_agent():
    agent = HealthAgent(
        id='agent',
        encoding=1,
        initial_health=0.45
    )
    assert isinstance(agent, GridWorldAgent)
    assert agent.initial_health == 0.45
    assert agent.configured

    with pytest.raises(AssertionError):
        agent = HealthAgent(
            id='agent',
            encoding=1,
            initial_health=2
        )


def test_attacking_agent():
    agent = AttackingAgent(
        id='agent',
        encoding=1,
        attack_range=3,
        attack_strength=0.6,
        attack_accuracy=0.95
    )
    assert isinstance(agent, ActingAgent)
    assert isinstance(agent, GridWorldAgent)
    assert agent.attack_range == 3
    assert agent.attack_strength == 0.6
    assert agent.attack_accuracy == 0.95

    with pytest.raises(AssertionError):
        agent = AttackingAgent(
            id='agent',
            encoding=1,
            attack_range=3.0,
            attack_strength=0.6,
            attack_accuracy=0.95
        )

    with pytest.raises(AssertionError):
        agent = AttackingAgent(
            id='agent',
            encoding=1,
            attack_range=3,
            attack_strength=2,
            attack_accuracy=0.95
        )

    with pytest.raises(AssertionError):
        agent = AttackingAgent(
            id='agent',
            encoding=1,
            attack_range=3,
            attack_strength=0.6,
            attack_accuracy=-0.3
        )




from gym.spaces import Discrete, MultiBinary
import numpy as np
import pytest

from abmarl.policies.q_table_policy import QTablePolicy, EpsilonSoftPolicy


class QPolicyTester(QTablePolicy):
    def reset(self):
        self.first_guess = True

    def compute_action(self, obs, **kwargs):
        pass

    def probability(self, obs, action, **kwargs):
        pass


def test_q_policy_init_and_properties():
    with pytest.raises(AssertionError):
        QPolicyTester(observation_space=Discrete(3), action_space=MultiBinary(4))
    with pytest.raises(AssertionError):
        QPolicyTester(observation_space=MultiBinary(2), action_space=Discrete(10))
    policy = QPolicyTester(observation_space=Discrete(3), action_space=(Discrete(10)))
    assert policy.q_table.shape == (3, 10)
    assert policy.action_space == Discrete(10)
    assert policy.observation_space == Discrete(3)
    policy = QPolicyTester(
        observation_space=Discrete(3),
        action_space=(Discrete(10)),
        q_table=np.random.randint(0, 5, (3, 10))
    )

    with pytest.raises(AssertionError):
        QPolicyTester(
            observation_space=Discrete(2),
            action_space=Discrete(10),
            q_table=np.random.normal(0, 1, (3, 10))
        )
    with pytest.raises(AssertionError):
        policy.q_table = Discrete(10)
    with pytest.raises(AssertionError):
        policy.action_space = MultiBinary(10)
    with pytest.raises(AssertionError):
        policy.observation_space = MultiBinary(10)


def test_q_policy_update():
    q = np.array([
        [ 0,  1,  2],
        [-1,  1, -2],
        [ 0,  1,  0],
        [ 2, -2, -2]])
    policy = QPolicyTester(
        observation_space=Discrete(q.shape[0]),
        action_space=Discrete(q.shape[1]),
        q_table=q
    )
    policy.update(0, 1, 3)
    policy.update(2, 2, -1)
    policy.update(3, 0, -2)
    np.testing.assert_array_equal(
        policy.q_table,
        np.array([
            [ 0,  3,  2],
            [-1,  1, -2],
            [ 0,  1, -1],
            [-2, -2, -2]
        ])
    )


def test_greedy_policy():
    q = np.array([
        [ 0,  1,  2],
        [-1,  1, -2],
        [ 0,  1,  0],
        [ 2, -2, -2]])
    policy = EpsilonSoftPolicy(
        observation_space=Discrete(q.shape[0]),
        action_space=Discrete(q.shape[1]),
        q_table=q,
        epsilon=0
    )
    assert policy.compute_action(0) == 2
    assert policy.probability(0, 0) == 0
    assert policy.probability(0, 1) == 0
    assert policy.probability(0, 2) == 1

    assert policy.compute_action(1) == 1
    assert policy.probability(1, 0) == 0
    assert policy.probability(1, 1) == 1
    assert policy.probability(1, 2) == 0

    assert policy.compute_action(2) == 1
    assert policy.probability(2, 0) == 0
    assert policy.probability(2, 1) == 1
    assert policy.probability(2, 2) == 0

    assert policy.compute_action(3) == 0
    assert policy.probability(3, 0) == 1
    assert policy.probability(3, 1) == 0
    assert policy.probability(3, 2) == 0


def test_epsilon_soft_init_and_build():
    policy = EpsilonSoftPolicy(
        observation_space=Discrete(3),
        action_space=Discrete(5),
        epsilon=0.6
    )
    assert policy.epsilon == 0.6

    policy = EpsilonSoftPolicy(
        observation_space=policy.observation_space,
        action_space=policy.action_space,
        q_table=policy._q_table,
        epsilon=0.2
    )
    assert policy.epsilon == 0.2

    with pytest.raises(AssertionError):
        EpsilonSoftPolicy(
            observation_space=Discrete(3),
            action_space=Discrete(5),
            epsilon=-0.6
        )

    with pytest.raises(AssertionError):
        EpsilonSoftPolicy(
            observation_space=policy.observation_space,
            action_space=policy.action_space,
            q_table=policy.q_table,
            epsilon=1.6
        )


def test_epsilon_soft_compute_action_and_probability():
    np.random.seed(24)
    q = np.array([
        [ 0,  1,  2],
        [-1,  1, -2],
        [ 0,  1,  0],
        [ 2, -2, -2]])
    policy = EpsilonSoftPolicy(
        observation_space=Discrete(q.shape[0]),
        action_space=Discrete(q.shape[1]),
        q_table=q,
        epsilon=0.5
    )
    assert policy.compute_action(0) == 2 # random
    assert policy.probability(0, 0) == 0.5 / 3
    assert policy.probability(0, 1) == 0.5 / 3
    assert policy.probability(0, 2) == 1 - 0.5 + 0.5 / 3

    assert policy.compute_action(1) == 1 # random
    assert policy.probability(1, 0) == 0.5 / 3
    assert policy.probability(1, 1) == 1 - 0.5 + 0.5 / 3
    assert policy.probability(1, 2) == 0.5 / 3

    assert policy.compute_action(2) == 1 # random
    assert policy.probability(2, 0) == 0.5 / 3
    assert policy.probability(2, 1) == 1 - 0.5 + 0.5 / 3
    assert policy.probability(2, 2) == 0.5 / 3

    assert policy.compute_action(3) == 0 # random
    assert policy.probability(3, 0) == 1 - 0.5 + 0.5 / 3
    assert policy.probability(3, 1) == 0.5 / 3
    assert policy.probability(3, 2) == 0.5 / 3




from abmarl.sim.gridworld.agent import HealthAgent
from abmarl.sim.gridworld.state import HealthState
from abmarl.sim.gridworld.done import ActiveDone, DoneBaseComponent
from abmarl.sim.gridworld.grid import Grid


def test_active_done():
    grid = Grid(2, 3)
    agents = {
        'agent0': HealthAgent(id='agent0', encoding=1),
        'agent1': HealthAgent(id='agent1', encoding=1),
        'agent2': HealthAgent(id='agent2', encoding=1),
    }

    health_state = HealthState(agents=agents, grid=grid)
    health_state.reset()

    active_done = ActiveDone(agents=agents, grid=grid)
    assert isinstance(active_done, DoneBaseComponent)

    assert agents['agent0'].active
    assert agents['agent1'].active
    assert agents['agent2'].active
    assert not active_done.get_done(agents['agent0'])
    assert not active_done.get_done(agents['agent1'])
    assert not active_done.get_done(agents['agent2'])
    assert not active_done.get_all_done()

    agents['agent0'].health = 0
    agents['agent1'].health = 0
    assert active_done.get_done(agents['agent0'])
    assert active_done.get_done(agents['agent1'])
    assert not active_done.get_done(agents['agent2'])
    assert not active_done.get_all_done()

    agents['agent2'].health = 0
    assert active_done.get_done(agents['agent2'])
    assert active_done.get_all_done()




import numpy as np

from abmarl.sim.gridworld.grid import Grid
from abmarl.sim.gridworld.state import PositionState, HealthState, StateBaseComponent
from abmarl.sim.gridworld.agent import HealthAgent, GridWorldAgent


def test_position_state():
    grid = Grid(3, 3)
    agents = {
        'agent0': GridWorldAgent(id='agent0', encoding=1, initial_position=np.array([0, 1])),
        'agent1': GridWorldAgent(id='agent1', encoding=1, initial_position=np.array([1, 2])),
        'agent2': GridWorldAgent(id='agent2', encoding=1, initial_position=np.array([2, 0]))
    }

    position_state = PositionState(grid=grid, agents=agents)
    assert isinstance(position_state, StateBaseComponent)
    position_state.reset()

    np.testing.assert_equal(agents['agent0'].position, np.array([0, 1]))
    np.testing.assert_equal(agents['agent1'].position, np.array([1, 2]))
    np.testing.assert_equal(agents['agent2'].position, np.array([2, 0]))
    assert grid[0, 1] == {'agent0': agents['agent0']}
    assert grid[1, 2] == {'agent1': agents['agent1']}
    assert grid[2, 0] == {'agent2': agents['agent2']}


def test_health_state():
    grid = Grid(3, 3)
    agents = {
        'agent0': HealthAgent(id='agent0', encoding=1, initial_health=0.24),
        'agent1': HealthAgent(id='agent1', encoding=1),
        'agent2': HealthAgent(id='agent2', encoding=1)
    }

    health_state = HealthState(agents=agents, grid=grid)
    assert isinstance(health_state, StateBaseComponent)
    health_state.reset()

    assert agents['agent0'].health == 0.24
    assert 0 <= agents['agent1'].health <= 1
    assert 0 <= agents['agent2'].health <= 1
    assert agents['agent0'].active
    assert agents['agent1'].active
    assert agents['agent2'].active




import numpy as np

from abmarl.sim.corridor import MultiCorridor
from abmarl.managers import AllStepManager
from abmarl.sim.wrappers import RavelDiscreteWrapper
from abmarl.policies.q_table_policy import EpsilonSoftPolicy
from abmarl.trainers.monte_carlo import OnPolicyMonteCarloTrainer


def test_epsilon_soft():
    np.random.seed(24)
    sim = AllStepManager(RavelDiscreteWrapper(MultiCorridor(num_agents=1)))
    ref_agent = sim.agents['agent0']
    policy = EpsilonSoftPolicy(
        action_space=ref_agent.action_space,
        observation_space=ref_agent.observation_space
    )
    trainer = OnPolicyMonteCarloTrainer(sim=sim, policy=policy)
    trainer.train(iterations=100, horizon=20)

    policy.epsilon = 0
    obs = sim.reset()
    for _ in range(10):
        actions = trainer.compute_actions(obs)
        obs, reward, done, info = sim.step(actions)
        if done['__all__']:
            break
    assert done['__all__']



import itertools

import numpy as np
from gym.spaces import Discrete, MultiDiscrete, MultiBinary, Box, Dict, Tuple

from .sar_wrapper import SARWrapper


def _ravel_helper(space, point):
    if isinstance(space, Discrete):
        return point, space.n
    if isinstance(space, MultiDiscrete):
        return np.ravel_multi_index(point, space.nvec), np.prod(space.nvec)
    if isinstance(space, MultiBinary):
        return np.ravel_multi_index(point, [2] * space.n), 2 ** space.n
    if isinstance(space, Box):
        space_helper = (space.high + 1 - space.low).flatten()
        return np.ravel_multi_index((point - space.low).flatten(), space_helper), \
            np.prod(space_helper)
    elif isinstance(space, Dict):
        discretized_values = []
        space_dims = []
        for key, space in space.spaces.items():
            discretized_value, space_dim = _ravel_helper(space, point[key])
            discretized_values.append(discretized_value)
            space_dims.append(space_dim)
        return _ravel_helper(MultiDiscrete(space_dims), discretized_values)
    elif isinstance(space, Tuple):
        discretized_values = []
        space_dims = []
        for point_part, space in zip(point, space.spaces):
            discretized_value, space_dim = _ravel_helper(space, point_part)
            discretized_values.append(discretized_value)
            space_dims.append(space_dim)
        return _ravel_helper(MultiDiscrete(space_dims), discretized_values)


def _nested_dim_helper(space):
    if isinstance(space, Discrete):
        return [space.n]
    elif isinstance(space, MultiDiscrete):
        return [np.prod(space.nvec)]
    elif isinstance(space, MultiBinary):
        return [2 ** space.n]
    if isinstance(space, Box):
        return [np.prod(space.high + 1 - space.low)]
    elif isinstance(space, Dict):
        return [np.prod([_nested_dim_helper(s) for s in space.spaces.values()])]
    elif isinstance(space, Tuple):
        return [np.prod([_nested_dim_helper(s) for s in space.spaces])]
    else:
        raise TypeError


def _nested_dim(space):
    """
    Return the total number of dimensions in the entire (nested) space.
    """
    if isinstance(space, Dict):
        return [
            *itertools.chain.from_iterable([_nested_dim_helper(s) for s in space.spaces.values()])
        ]
    elif isinstance(space, Tuple):
        return [*itertools.chain.from_iterable([_nested_dim_helper(s) for s in space.spaces])]
    else:
        return _nested_dim_helper(space)


def ravel(space, point):
    """
    Ravel point in space to a single discrete value.
    """
    return _ravel_helper(space, point)[0]


def unravel(space, point):
    """
    Unravel a single discrete point to a value in the space.
    """
    if isinstance(space, Discrete):
        return point
    if isinstance(space, MultiDiscrete):
        return [*np.unravel_index(point, space.nvec)]
    if isinstance(space, MultiBinary):
        return [*np.unravel_index(point, [2] * space.n)]
    if isinstance(space, Box):
        space_helper = (space.high + 1 - space.low).flatten()
        return np.reshape(np.unravel_index(point, space_helper), space.shape) + space.low
    elif isinstance(space, Dict):
        dims = _nested_dim(space)
        unravelled_point = unravel(MultiDiscrete(dims), point)
        output = {}
        for i, (key, value) in enumerate(space.spaces.items()):
            output[key] = unravel(value, unravelled_point[i])
        return output
    elif isinstance(space, Tuple):
        dims = _nested_dim(space)
        unravelled_point = unravel(MultiDiscrete(dims), point)
        output = []
        for i, value in enumerate(space.spaces):
            output.append(unravel(value, unravelled_point[i]))
        return tuple(output)


def ravel_space(space):
    """
    Convert the space into a Discrete space.
    """
    dims = _nested_dim_helper(space)
    return Discrete(dims[0])


def _isbounded(space):
    """
    Gym Box converts np.inf to min and max values for integer types. As a result,
    Box.is_bounded doesn't work because it checks for inf, not for min/max values
    of that dtype. This function checks for min/max values of the dtype.
    """
    return space.is_bounded() and \
        not (space.low == np.iinfo(space.dtype).min).any() and \
        not (space.low == np.iinfo(space.dtype).max).any() and \
        not (space.high == np.iinfo(space.dtype).min).any() and \
        not (space.high == np.iinfo(space.dtype).max).any()


def check_space(space):
    """
    Ensure that the space is of type that can be ravelled to discrete value.
    """
    if isinstance(space, Discrete) or isinstance(space, MultiDiscrete) or \
            isinstance(space, MultiBinary):
        return True
    elif isinstance(space, Box) and np.issubdtype(space, int) and _isbounded(space):
        return True
    elif isinstance(space, Dict):
        return all([check_space(sub_space) for sub_space in space.spaces.values()])
    elif isinstance(space, Tuple):
        return all([check_space(sub_space) for sub_space in space.spaces])
    else:
        return False


class RavelDiscreteWrapper(SARWrapper):
    """
    Convert Discrete, MultiBinary, MultiDiscrete, bounded integer Box, and any nesting of these
    observations and actions into Discrete observations and actions by "ravelling" their values
    according to numpy's ravel_mult_index function. Thus, observations and actions that are
    represented by arrays are converted into unique numbers. This is useful for building Q
    tables where each observation and action is a row and column of the Q table, respectively.
    """
    def __init__(self, sim):
        super().__init__(sim)
        for agent_id, wrapped_agent in self.agents.items():
            assert check_space(wrapped_agent.observation_space), \
                f"{agent_id}: observation must be discretizable."
            assert check_space(wrapped_agent.action_space), \
                f"{agent_id} action must be discretizable."
            self.agents[agent_id].observation_space = ravel_space(wrapped_agent.observation_space)
            self.agents[agent_id].action_space = ravel_space(wrapped_agent.action_space)

    def wrap_observation(self, from_agent, observation):
        return ravel(from_agent.observation_space, observation)

    def unwrap_observation(self, from_agent, observation):
        return unravel(from_agent.observation_space, observation)

    def wrap_action(self, from_agent, action):
        return unravel(from_agent.action_space, action)

    def unwrap_action(self, from_agent, action):
        return ravel(from_agent.action_space, action)
